from abc import ABC, abstractmethod
from ..model import (
    EngineResponse,
    ModelAlreadyLoadedException,
    TextGenerationVendor,
    TokenizerAlreadyLoadedException,
    TokenizerNotSupportedException,
)
from ..model.entities import (
    EngineSettings,
    Input,
    ModelConfig,
    SentenceTransformerModelConfig,
    TokenizerConfig,
)
from contextlib import ExitStack
from importlib.util import find_spec
from logging import ERROR, Logger, getLogger
from torch import cuda
from torch.backends import mps
from transformers import logging as transformers_logging
from transformers.utils.logging import disable_progress_bar, enable_progress_bar
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)
from typing import Any, Type, Union


class Engine(ABC):
    _device: str
    _logger: Logger
    _model_id: str
    _settings: EngineSettings
    _transformers_logging_logger: Logger
    _transformers_logging_level: int
    _loaded_model: bool = False
    _loaded_tokenizer: bool = False
    _tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast] | None = (
        None
    )
    _model: Union[PreTrainedModel, TextGenerationVendor] | None = None
    _config: Union[ModelConfig, SentenceTransformerModelConfig] | None = None
    _tokenizer_config: TokenizerConfig | None = None
    _exit_stack: ExitStack = ExitStack()

    def __init__(
        self,
        model_id: str,
        settings: EngineSettings | None = None,
        logger: Logger | None = None,
    ):
        assert model_id
        self._logger = logger if logger else getLogger()
        self._model_id = model_id
        self._settings = settings if settings else EngineSettings()
        self._device = (
            self._settings.device
            if self._settings.device
            else Engine.get_default_device()
        )
        self._transformers_logging_logger = (
            transformers_logging.get_logger()
            if self._settings.change_transformers_logging_level
            else None
        )
        self._transformers_logging_level = (
            self._transformers_logging_logger.level
            if self._settings.change_transformers_logging_level
            else None
        )

        auto_load_tokenizer = (
            self.uses_tokenizer and self._settings.auto_load_tokenizer
        )
        if self._settings.auto_load_model or auto_load_tokenizer:
            self._load(
                load_tokenizer=auto_load_tokenizer,
                tokenizer_name_or_path=self._settings.tokenizer_name_or_path
                if auto_load_tokenizer
                else None,
            )

    @property
    def uses_tokenizer(self) -> bool:
        return False

    @property
    def config(
        self,
    ) -> Union[ModelConfig, SentenceTransformerModelConfig] | None:
        return self._config

    @property
    def model(self) -> PreTrainedModel | None:
        return self._model

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def tokenizer_config(self) -> TokenizerConfig | None:
        return self._tokenizer_config

    @property
    def tokenizer(
        self,
    ) -> Union[PreTrainedTokenizer, PreTrainedTokenizerFast] | None:
        return self._tokenizer

    @abstractmethod
    async def __call__(self, input: Input, **kwargs) -> EngineResponse:
        raise NotImplementedError()

    @abstractmethod
    def _load_model(self) -> Union[PreTrainedModel, TextGenerationVendor]:
        raise NotImplementedError()

    def _load_tokenizer_with_tokens(
        self, tokenizer_name_or_path: str | None, use_fast: bool = True
    ) -> Union[PreTrainedTokenizer, PreTrainedTokenizerFast]:
        raise (
            TokenizerNotSupportedException()
            if not self.uses_tokenizer()
            else NotImplementedError()
        )

    def __enter__(self):
        _l = self._log
        if (
            self._transformers_logging_logger
            and self._transformers_logging_level != ERROR
        ):
            transformers_logging.set_verbosity_error()
            _l(
                "Changed transformers logging level from "
                f"{self._transformers_logging_level} to "
                f"{self._transformers_logging_logger.level}"
            )
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any | None,
    ):
        _l = self._log
        if (
            self._transformers_logging_logger
            and self._transformers_logging_level
            and self._transformers_logging_level != ERROR
        ):
            self._transformers_logging_logger.setLevel(
                self._transformers_logging_level
            )
            _l(
                f"Restored transformers logging level to "
                f"{self._transformers_logging_level}"
            )
        return self._exit_stack.__exit__(exc_type, exc_value, traceback)

    def _load(
        self, *args, load_tokenizer: bool, tokenizer_name_or_path: str | None
    ) -> None:
        if (
            self._settings.auto_load_model
            and hasattr(self, "_loaded_model")
            and self._loaded_model
        ):
            raise ModelAlreadyLoadedException()
        elif (
            load_tokenizer
            and hasattr(self, "_loaded_tokenizer")
            and self._loaded_tokenizer
        ):
            raise TokenizerAlreadyLoadedException()

        _l = self._log

        if self._settings.disable_loading_progress_bar:
            disable_progress_bar()

        if load_tokenizer:
            _l(f"Loading tokenizer {tokenizer_name_or_path or self._model_id}")
            self._tokenizer = self._load_tokenizer_with_tokens(
                tokenizer_name_or_path or self._model_id, use_fast=True
            )
            assert isinstance(
                self._tokenizer, PreTrainedTokenizer
            ) or isinstance(self._tokenizer, PreTrainedTokenizerFast), (
                "Unexpected pretrained tokenizer type: "
                + f"{type(self._tokenizer)}"
            )
            _l(f"Loaded tokenizer {self._tokenizer.name_or_path}")

            self._loaded_tokenizer = True

        is_sentence_transformer = False
        if self._settings.auto_load_model:
            _l(
                f"Loading pretrained model {self._model_id} from "
                f"cache {self._settings.cache_dir}"
            )
            self._model = self._load_model()

            is_mlx = False
            is_sentence_transformer = False

            if not isinstance(self._model, PreTrainedModel) and not isinstance(
                self._model, TextGenerationVendor
            ):
                if find_spec("mlx"):
                    from mlx.nn import Module

                    is_mlx = isinstance(self._model, Module)

                if find_spec("sentence_transformers"):
                    from sentence_transformers import SentenceTransformer

                    is_sentence_transformer = isinstance(
                        self._model, SentenceTransformer
                    )

            assert (
                isinstance(self._model, PreTrainedModel)
                or isinstance(self._model, TextGenerationVendor)
                or is_mlx
                or is_sentence_transformer
            ), f"Unexpected pretrained model type: {type(self._model)}"

            _l(
                f"Loaded pretrained model {self._model_id} from "
                f"cache {self._settings.cache_dir}"
            )

            if self._settings.enable_eval:
                _l(f"Setting model {self._model_id} in eval mode")
                self._model.eval()

            if self._tokenizer and (
                self._settings.tokens or self._settings.special_tokens
            ):
                total_tokens = len(self._tokenizer)
                _l(
                    f"Resizing embedding matrix to {total_tokens} tokens "
                    f"for model {self._model_id}"
                )
                self._model.resize_token_embeddings(total_tokens)
                _l(
                    f"Resized embedding matrix to {total_tokens} tokens "
                    f"for model {self._model_id}"
                )

            self._loaded_model = True

        if self._model and not self._config:
            config: (
                Union[ModelConfig, SentenceTransformerModelConfig] | None
            ) = None
            mc = (
                self._model.config
                if hasattr(self._model, "config")
                else self._model[0].auto_model.config
                if is_sentence_transformer
                else None
            )

            if mc:
                config = ModelConfig(
                    architectures=mc.architectures,
                    attribute_map=mc.attribute_map,
                    bos_token_id=mc.bos_token_id,
                    bos_token=self._tokenizer.decode(mc.bos_token_id)
                    if self._tokenizer and mc.bos_token_id
                    else None,
                    decoder_start_token_id=mc.decoder_start_token_id,
                    eos_token_id=mc.eos_token_id,
                    eos_token=self._tokenizer.decode(mc.eos_token_id)
                    if self._tokenizer and mc.eos_token_id
                    else None,
                    finetuning_task=mc.finetuning_task,
                    hidden_size=mc.hidden_size
                    if hasattr(mc, "hidden_size")
                    else None,
                    hidden_sizes=mc.hidden_sizes
                    if hasattr(mc, "hidden_sizes")
                    else None,
                    keys_to_ignore_at_inference=mc.keys_to_ignore_at_inference
                    if hasattr(mc, "keys_to_ignore_at_inference")
                    else None,
                    loss_type=mc.loss_type
                    if hasattr(mc, "loss_type")
                    else None,
                    max_position_embeddings=mc.max_position_embeddings
                    if hasattr(mc, "max_position_embeddings")
                    else None,
                    model_type=mc.model_type,
                    num_attention_heads=mc.num_attention_heads
                    if hasattr(mc, "num_attention_heads")
                    else None,
                    num_hidden_layers=mc.num_hidden_layers
                    if hasattr(mc, "num_hidden_layers")
                    else None,
                    num_labels=mc.num_labels,
                    output_attentions=mc.output_attentions,
                    output_hidden_states=mc.output_hidden_states,
                    pad_token_id=mc.pad_token_id,
                    pad_token=self._tokenizer.decode(mc.pad_token_id)
                    if self._tokenizer and mc.pad_token_id
                    else None,
                    prefix=mc.prefix,
                    sep_token_id=mc.sep_token_id,
                    sep_token=self._tokenizer.decode(mc.sep_token_id)
                    if self._tokenizer and mc.sep_token_id
                    else None,
                    state_size=len(self._model.state_dict().keys())
                    if self._model.state_dict
                    else 0,
                    task_specific_params=mc.task_specific_params,
                    torch_dtype=str(mc.torch_dtype),
                    vocab_size=mc.vocab_size
                    if hasattr(mc, "vocab_size")
                    else None,
                    tokenizer_class=mc.tokenizer_class,
                )

            if is_sentence_transformer and config:
                config = SentenceTransformerModelConfig(
                    backend=self._model.backend,
                    similarity_function=self._model.similarity_fn_name,
                    truncate_dimension=self._model.truncate_dim,
                    transformer_model_config=config,
                )

            self._config = config

        if self._tokenizer and not self._tokenizer_config:
            self._tokenizer_config = TokenizerConfig(
                name_or_path=self._tokenizer.name_or_path,
                tokens=self._settings.tokens,
                special_tokens=self._tokenizer.all_special_tokens,
                tokenizer_model_max_length=self._tokenizer.model_max_length,
                fast=isinstance(self._tokenizer, PreTrainedTokenizerFast),
            )

        if self._settings.disable_loading_progress_bar:
            enable_progress_bar()

    @staticmethod
    def get_default_device() -> str:
        return (
            "cuda"
            if cuda.is_available()
            else "mps"
            if mps.is_available()
            else "cpu"
        )

    def _log(self, message: str) -> None:
        self._logger.debug(f"<{self._model_id}> {message}")



from __future__ import annotations
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..flow.flow import Flow


class Node:
    def __init__(
        self,
        name: str,
        label: Optional[str] = None,
        shape: Optional[str] = None,
        input_schema: Optional[type] = None,
        output_schema: Optional[type] = None,
        func: Optional[Callable[..., Any]] = None,
        subgraph: Optional[Flow] = None,
    ) -> None:
        self.name: str = name
        self.label: str = label or name
        self.shape: Optional[str] = shape
        self.input_schema: Optional[type] = input_schema
        self.output_schema: Optional[type] = output_schema
        self.func: Optional[Callable[..., Any]] = func
        self.subgraph: Optional[Flow] = subgraph

    def execute(self, inputs: Dict[str, Any]) -> Any:
        # Delegate to subgraph if present
        if self.subgraph is not None:
            initial = (
                next(iter(inputs.values())) if len(inputs) == 1 else inputs
            )
            result = self.subgraph.execute(
                initial_node=None, initial_data=initial
            )
            if self.output_schema and not isinstance(
                result, self.output_schema
            ):
                raise TypeError(
                    f"{self.name} output {result!r} not {self.output_schema}"
                )
            return result

        # Validate input schema
        if self.input_schema:
            if isinstance(self.input_schema, type):
                if isinstance(inputs, dict) and len(inputs) == 1:
                    val = next(iter(inputs.values()))
                    if not isinstance(val, self.input_schema):
                        raise TypeError(
                            f"{self.name} input {val!r} not {self.input_schema}"
                        )
                elif not isinstance(inputs, self.input_schema):
                    raise TypeError(
                        f"{self.name} input {inputs!r} not {self.input_schema}"
                    )

        # Compute output
        if callable(self.func):
            try:
                output = self.func(inputs)
            except TypeError:
                output = self.func(*inputs.values())  # type: ignore
        else:
            if not inputs:
                output = None
            elif len(inputs) == 1:
                output = next(iter(inputs.values()))
            else:
                output = inputs

        # Validate output schema
        if self.output_schema and output is not None:
            if not isinstance(output, self.output_schema):
                raise TypeError(
                    f"{self.name} output {output!r} not {self.output_schema}"
                )

        return output

    def __repr__(self) -> str:
        return f"<Node {self.name}>"



from abc import ABC, abstractmethod
from ..model.entities import (
    GenerationSettings,
    ImageEntity,
    Message,
    Token,
    TokenDetail,
)
from ..tool.manager import ToolManager
from io import StringIO
from inspect import iscoroutine
from json import loads, JSONDecodeError
from re import compile, DOTALL, Pattern
from numpy import ndarray
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Literal,
    Generator,
    Union,
    TypedDict,
)

OutputGenerator = AsyncGenerator[Token | TokenDetail | str, None]
OutputFunction = Callable[..., OutputGenerator | str]

TemplateMessageRole = Literal["assistant", "system", "tool", "user"]


class InvalidJsonResponseException(Exception):
    pass


class ModelAlreadyLoadedException(Exception):
    pass


class TokenizerAlreadyLoadedException(Exception):
    pass


class TokenizerNotSupportedException(Exception):
    pass


class TemplateMessage(TypedDict):
    role: TemplateMessageRole
    content: str


class TextGenerationResponse:
    _json_patterns: list[Pattern] = [
        # Markdown code fence with explicit json tag
        compile(r"```json\s*(\{.*?\})\s*```", DOTALL),
        # Any markdown code fence possibly with a language specifier
        compile(r"```(?:\w+)?\s*(\{.*?\})\s*```", DOTALL),
        # Generic JSON-like pattern
        compile(r"(\{.*\})", DOTALL),
    ]
    _output_fn: OutputFunction
    _input_token_count: int = 0
    _output: OutputGenerator | None = None
    _buffer: StringIO = StringIO()
    _on_consumed: Callable[[], Awaitable[None] | None] | None = None
    _consumed: bool = False

    def __init__(
        self,
        output_fn: OutputFunction,
        *args,
        use_async_generator: bool,
        **kwargs,
    ):
        self._args = args
        self._kwargs = kwargs
        self._output_fn = output_fn
        self._use_async_generator = use_async_generator

        if "inputs" in self._kwargs:
            inputs = self._kwargs["inputs"]
            self._input_token_count = (
                len(inputs["input_ids"][0])
                if inputs and "input_ids" in inputs
                else 0
            )

    def add_done_callback(
        self, callback: Callable[[], Awaitable[None] | None]
    ) -> None:
        self._on_consumed = callback

    @property
    def input_token_count(self) -> int:
        return self._input_token_count

    async def _trigger_consumed(self) -> None:
        if self._consumed:
            return
        self._consumed = True
        if self._on_consumed:
            result = self._on_consumed()
            if iscoroutine(result):
                await result

    def __aiter__(self):
        # Create a fresh async generator each time we start iterating
        self._output = self._output_fn(*self._args, **self._kwargs)
        return self

    async def __anext__(self) -> Token | TokenDetail | str:
        try:
            token = await self._output.__anext__()
        except StopAsyncIteration:
            await self._trigger_consumed()
            raise
        self._buffer.write(token if isinstance(token, str) else token.token)
        return token

    async def to_str(self) -> str:
        if not self._use_async_generator:
            result = self._output_fn(*self._args, **self._kwargs)
            self._buffer.write(result)
            await self._trigger_consumed()
            return result

        # Ensure buffer is filled, wether we were already iterating or not
        if not self._output:
            self.__aiter__()

        async for token in self._output:
            self._buffer.write(token)

        await self._trigger_consumed()
        return self._buffer.getvalue()

    async def to_json(self) -> str:
        text = await self.to_str()
        assert text
        for pattern in self._json_patterns:
            match = pattern.search(text)
            if match:
                json_str = match.group(1)
                try:
                    loads(json_str)
                    return json_str
                except JSONDecodeError:
                    continue
        raise InvalidJsonResponseException(text)

    async def to(self, entity_class: type) -> any:
        json = await self.to_json()
        data = loads(json)
        return entity_class(**data)


class TextGenerationStream(AsyncIterator[Token | TokenDetail | str], ABC):
    _generator: AsyncGenerator | None = None

    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    async def __anext__(self) -> Token | TokenDetail | str:
        raise NotImplementedError()

    def __aiter__(self):
        assert self._generator
        return self


class TextGenerationVendor(ABC):
    async def __call__(
        self,
        model_id: str,
        messages: list[Message],
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
        use_async_generator: bool = True,
    ) -> TextGenerationStream:
        raise NotImplementedError()

    def _system_prompt(self, messages: list[Message]) -> str | None:
        return next(
            (
                message.content
                for message in messages
                if message.role == "system"
            ),
            None,
        )

    def _template_messages(
        self,
        messages: list[Message],
        exclude_roles: list[TemplateMessageRole] | None = None,
    ) -> list[TemplateMessage]:
        return [
            {"role": message.role, "content": message.content}
            for message in messages
            if not exclude_roles or message.role not in exclude_roles
        ]


class TextGenerationVendorStream(TextGenerationStream):
    _generator: AsyncGenerator

    def __init__(self, generator: AsyncGenerator):
        self._generator = generator

    def __call__(self, *args, **kwargs):
        return self.__aiter__()

    def __aiter__(self):
        assert self._generator
        return self


EngineResponse = Union[
    TextGenerationResponse,
    TextGenerationVendorStream,
    Generator[str, None, None],
    Generator[Union[Token, TokenDetail], None, None],
    ImageEntity,
    list[ImageEntity],
    list[str],
    dict[str, str],
    ndarray,
    str,
]



from abc import ABC, abstractmethod
from ..agent import Specification
from ..memory.manager import MemoryManager
from ..model import TextGenerationResponse
from ..model.engine import Engine
from ..model.entities import (
    EngineMessage,
    GenerationSettings,
    Input,
    Message,
    MessageRole,
)
from ..model.nlp.text.vendor import TextGenerationVendorModel
from ..tool.manager import ToolManager
from ..event import Event, EventType
from ..event.manager import EventManager
from dataclasses import replace
from typing import Any, Optional, Union, Tuple
from uuid import UUID, uuid4


class EngineAgent(ABC):
    _id: UUID
    _name: Optional[str]
    _model: Engine
    _memory: MemoryManager
    _tool: ToolManager
    _event_manager: EventManager
    _last_output: Optional[TextGenerationResponse] = None
    _last_prompt: Optional[Tuple[Input, Optional[str]]] = None

    @abstractmethod
    def _prepare_call(
        self, specification: Specification, input: str, **kwargs: Any
    ) -> Any:
        raise NotImplementedError()

    @property
    def memory(self) -> MemoryManager:
        return self._memory

    @property
    def engine(self) -> Engine:
        return self._model

    @property
    def output(self) -> Optional[TextGenerationResponse]:
        return self._last_output

    async def input_token_count(self) -> Optional[int]:
        if not self._last_prompt:
            return None
        await self._event_manager.trigger(
            Event(type=EventType.INPUT_TOKEN_COUNT_BEFORE)
        )
        count = self._model.input_token_count(
            self._last_prompt[0], system_prompt=self._last_prompt[1]
        )
        await self._event_manager.trigger(
            Event(
                type=EventType.INPUT_TOKEN_COUNT_AFTER, payload={"count": count}
            )
        )
        return count

    def __init__(
        self,
        model: Engine,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        *args,
        name: Optional[str] = None,
        id: Optional[UUID] = None,
    ):
        self._id = id or uuid4()
        self._name = name
        self._model = model
        self._memory = memory
        self._tool = tool
        self._event_manager = event_manager

    async def __call__(
        self, specification: Specification, input: str, **kwargs
    ) -> Union[TextGenerationResponse, str]:
        await self._event_manager.trigger(
            Event(type=EventType.CALL_PREPARE_BEFORE)
        )
        run_args = self._prepare_call(specification, input, **kwargs)
        await self._event_manager.trigger(
            Event(type=EventType.CALL_PREPARE_AFTER)
        )
        return await self._run(input, **run_args)

    async def _run(
        self,
        input: str,
        *args,
        settings: Optional[GenerationSettings] = None,
        system_prompt: Optional[str] = None,
        skip_special_tokens=True,
        **kwargs,
    ) -> Union[TextGenerationResponse]:
        # Process settings
        if settings and kwargs:
            settings = replace(settings, **kwargs)
        elif not settings:
            kwargs.setdefault("temperature", None)
            kwargs.setdefault("do_sample", False)
            settings = GenerationSettings(**kwargs)
        assert settings

        # Prepare memory
        assert (
            not self._memory.has_recent_message
            or self._memory.recent_message is not None
        )

        # Should always be stored, with or without memory
        self._last_prompt = (input, system_prompt)

        # Transform input (by adding memory, if necessary)
        if (
            self._memory.has_permanent_message
            or self._memory.has_recent_message
        ) and isinstance(input, Message):
            previous_message: Optional[Message] = None
            new_message: Message = input

            # Handle last message if not already consumed
            previous_output = self._last_output
            if previous_output and isinstance(
                previous_output, TextGenerationResponse
            ):
                previous_message = Message(
                    role=MessageRole.ASSISTANT,
                    content=await previous_output.to_str(),
                )

            # Append messages
            if previous_message:
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_APPEND_BEFORE,
                        payload={"message": previous_message},
                    )
                )
                await self._memory.append_message(
                    EngineMessage(
                        agent_id=self._id,
                        model_id=self._model.model_id,
                        message=previous_message,
                    )
                )
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_APPEND_AFTER,
                        payload={"message": previous_message},
                    )
                )
            await self._event_manager.trigger(
                Event(
                    type=EventType.MEMORY_APPEND_BEFORE,
                    payload={"message": new_message},
                )
            )
            await self._memory.append_message(
                EngineMessage(
                    agent_id=self._id,
                    model_id=self._model.model_id,
                    message=new_message,
                )
            )
            await self._event_manager.trigger(
                Event(
                    type=EventType.MEMORY_APPEND_AFTER,
                    payload={"message": new_message},
                )
            )

            # Make recent memory the new model input
            input = [rm.message for rm in self._memory.recent_messages]

        # Have model generate output from input

        model_settings = dict(
            system_prompt=system_prompt, settings=settings, tool=self._tool
        )
        if not isinstance(self._model, TextGenerationVendorModel):
            model_settings["skip_special_tokens"] = skip_special_tokens

        await self._event_manager.trigger(
            Event(type=EventType.MODEL_EXECUTE_BEFORE)
        )
        output = await self._model(input, **model_settings)
        await self._event_manager.trigger(
            Event(type=EventType.MODEL_EXECUTE_AFTER)
        )

        # Update memory
        if self._memory.has_recent_message:
            self._last_output = output

        return output



try:
    from keyring import get_keyring
    from keyring.backend import KeyringBackend
except Exception:  # pragma Module may not be installed
    get_keyring = None  # type: ignore
    KeyringBackend = object  # type: ignore


class KeyringSecrets:
    _SERVICE = "avalan"

    def __init__(self, ring: KeyringBackend | None = None):
        if ring is None and get_keyring:
            ring = get_keyring()
        self._ring = ring

    def read(self, key: str) -> str | None:
        assert self._ring, "keyring package not installed"
        return self._ring.get_password(self._SERVICE, key)

    def write(self, key: str, secret: str) -> None:
        assert self._ring, "keyring package not installed"
        self._ring.set_password(self._SERVICE, key, secret)

    def delete(self, key: str) -> None:
        assert self._ring, "keyring package not installed"
        try:
            self._ring.delete_password(self._SERVICE, key)
        except Exception:
            pass



from ... import name, version
from ...model.entities import (
    Model,
    HubCacheDeletion,
    HubCacheFile,
    HubCache,
    User,
)
from ...model.hubs import HubAccessDeniedException
from datetime import datetime
from huggingface_hub import HfApi, login, ModelInfo, scan_cache_dir
from huggingface_hub.errors import GatedRepoError
from logging import Logger
from os.path import expanduser
from os import getenv
from tqdm import tqdm
from typing import Callable, Iterable, Type
from urllib.parse import urlparse


class HuggingfaceHub:
    DEFAULT_ENDPOINT: str = "https://huggingface.co"
    DEFAULT_CACHE_DIR: str = expanduser(
        getenv("HF_HUB_CACHE") or "~/.cache/huggingface/hub"
    )
    _access_token: str
    _cache_dir: str
    _domain: str
    _hf: HfApi
    _logger: Logger

    @property
    def cache_dir(self) -> str:
        return self._cache_dir

    @property
    def domain(self) -> str:
        return self._domain

    def __init__(
        self,
        access_token: str,
        cache_dir: str,
        logger: Logger,
        endpoint: str = DEFAULT_ENDPOINT,
    ) -> None:
        assert access_token and cache_dir
        self._access_token = access_token
        self._hf = HfApi(
            endpoint=endpoint,
            token=access_token,
            library_name=name(),
            library_version=version(),
        )
        self._cache_dir = expanduser(cache_dir)
        self._domain = urlparse(endpoint).netloc
        self._logger = logger

    def cache_delete(
        self, model_id: str, revisions: list[str] | None = None
    ) -> (HubCacheDeletion | None, Callable[[], None] | None):
        scan_results = scan_cache_dir(self._cache_dir)
        delete_revisions = [
            revision.commit_hash
            for info in scan_results.repos
            if info.repo_id == model_id
            for revision in info.revisions
            if not revisions
            or any(revision.commit_hash.startswith(r) for r in revisions)
        ]
        if not delete_revisions:
            return (None, None)

        strategy = scan_results.delete_revisions(*delete_revisions)
        cache_deletion = HubCacheDeletion(
            model_id=model_id,
            revisions=delete_revisions,
            deletable_size_on_disk=strategy.expected_freed_size,
            deletable_blobs=[str(p) for p in strategy.blobs],
            deletable_refs=[str(p) for p in strategy.refs],
            deletable_repos=[str(p) for p in strategy.repos],
            deletable_snapshots=[str(p) for p in strategy.snapshots],
        )
        return cache_deletion, lambda: strategy.execute()

    def cache_scan(
        self, sort_models_by_size: bool = True, sort_files_by_size: bool = True
    ) -> list[HubCache]:
        scan_results = scan_cache_dir(self._cache_dir)
        model_caches = sorted(
            [
                HubCache(
                    model_id=info.repo_id,
                    path=str(info.repo_path),
                    size_on_disk=info.size_on_disk,
                    files={
                        revision.commit_hash: sorted(
                            [
                                HubCacheFile(
                                    name=rfile.file_name,
                                    path=str(rfile.file_path),
                                    size_on_disk=rfile.size_on_disk,
                                    last_accessed=datetime.fromtimestamp(
                                        rfile.blob_last_accessed
                                    ),
                                    last_modified=datetime.fromtimestamp(
                                        rfile.blob_last_modified
                                    ),
                                )
                                for rfile in revision.files
                            ],
                            key=lambda f: f.size_on_disk
                            if sort_files_by_size
                            else f.name,
                            reverse=sort_files_by_size,
                        )
                        for revision in info.revisions
                    },
                    revisions=[r.commit_hash for r in info.revisions],
                    total_files=info.nb_files,
                    total_revisions=len(info.revisions),
                )
                for info in scan_results.repos
            ],
            key=lambda m: m.size_on_disk if sort_models_by_size else m.name,
            reverse=sort_models_by_size,
        )
        return model_caches

    def can_access(self, model_id: str) -> bool:
        try:
            self._hf.auth_check(model_id)
        except GatedRepoError:
            return False
        return True

    def download(
        self,
        model_id: str,
        tqdm_class: Type[tqdm] | Callable[..., tqdm] | None = None,
    ) -> str:
        try:
            path = self._hf.snapshot_download(
                model_id,
                cache_dir=self._cache_dir,
                tqdm_class=tqdm_class,
                force_download=False,
            )
            return path
        except GatedRepoError as e:
            raise HubAccessDeniedException(e)

    def download_all(self, model_id: str) -> list[str]:
        files = self._hf.list_repo_files(model_id)
        for file in files:
            self._hf.hf_hub_download(
                model_id, file, cache_dir=self._cache_dir, force_download=False
            )
        return files

    def model(self, model_id: str) -> Model:
        model_info = self._hf.model_info(model_id)
        return HuggingfaceHub._model(model_info)

    def model_url(self, model_id: str) -> str:
        return f"https://huggingface.co/{model_id}"

    def models(
        self,
        filter: str | None = None,
        name: str | None = None,
        search: str | None = None,
        limit: int | None = None,
    ) -> Iterable[Model]:
        yield from (
            HuggingfaceHub._model(model_info)
            for model_info in self._hf.list_models(
                model_name=name,
                filter=filter,
                search=search,
                limit=limit,
                full=True,
            )
        )

    def login(self) -> None:
        login(self._access_token)

    def user(self) -> User:
        user_result = self._hf.whoami()
        return User(
            name=user_result["name"],
            full_name=user_result["fullname"],
            access_token_name=user_result["auth"]["accessToken"]["displayName"],
        )

    @staticmethod
    def _model(model_info: ModelInfo) -> Model:
        model = Model(
            id=model_info.id,
            parameters=model_info.safetensors.total
            if model_info.safetensors
            else None,
            parameter_types=list(model_info.safetensors.parameters.keys())
            if model_info.safetensors and model_info.safetensors.parameters
            else None,
            inference=model_info.inference,
            library_name=model_info.library_name
            if model_info.library_name
            else model_info.card_data["library_name"]
            if model_info.card_data and "library_name" in model_info.card_data
            else None,
            license=model_info.card_data["license"]
            if model_info.card_data and "license" in model_info.card_data
            else None,
            pipeline_tag=model_info.pipeline_tag,
            tags=model_info.tags,
            architectures=model_info.config["architectures"]
            if model_info.config and "architectures" in model_info.config
            else None,
            model_type=model_info.config["model_type"]
            if model_info.config and "model_type" in model_info.config
            else None,
            auto_model=model_info.transformers_info["auto_model"]
            if model_info.transformers_info
            and "auto_model" in model_info.transformers_info
            else None,
            processor=model_info.transformers_info["processor"]
            if model_info.transformers_info
            and "processor" in model_info.transformers_info
            else None,
            gated=model_info.gated,
            private=model_info.private,
            disabled=model_info.disabled,
            last_downloads=model_info.downloads,
            downloads=model_info.downloads_all_time or model_info.downloads,
            likes=model_info.likes,
            ranking=model_info.trending_score,
            author=model_info.author,
            created_at=model_info.created_at,
            updated_at=model_info.last_modified or model_info.created_at,
        )
        return model



from ..flow.node import Node
from typing import Any, Callable, Optional


class Connection:
    def __init__(
        self,
        src: Node,
        dest: Node,
        label: Optional[str] = None,
        conditions: Optional[list[Callable[[Any], bool]]] = None,
        filters: Optional[list[Callable[[Any], Any]]] = None,
    ) -> None:
        self.src: Node = src
        self.dest: Node = dest
        self.label: Optional[str] = label
        self.conditions: list[Callable[[Any], bool]] = conditions or []
        self.filters: list[Callable[[Any], Any]] = filters or []

    def check_conditions(self, data: Any) -> bool:
        return all(cond(data) for cond in self.conditions)

    def apply_filters(self, data: Any) -> Any:
        for f in self.filters:
            data = f(data)
        return data

    def __repr__(self) -> str:
        return f"<Conn {self.src.name}->{self.dest.name}>"



from ...agent.orchestrator import Orchestrator
from ...server.entities import (
    ChatCompletionChoice,
    ChatCompletionChunk,
    ChatCompletionChunkChoice,
    ChatCompletionChunkChoiceDelta,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatMessage,
    ChatCompletionUsage,
)
from ...model.entities import GenerationSettings, Message, MessageRole
from fastapi import APIRouter, Depends, Request
from fastapi.responses import StreamingResponse
from time import time

router = APIRouter(
    prefix="/chat",
    tags=["completions"],
)


def dependency_get_orchestrator(request: Request) -> Orchestrator:
    return request.app.state.orchestrator


@router.post("/completions", response_model=ChatCompletionResponse)
async def create_chat_completion(
    request: ChatCompletionRequest,
    orchestrator: Orchestrator = Depends(dependency_get_orchestrator),
):
    assert orchestrator and isinstance(orchestrator, Orchestrator)
    assert request and request.messages

    # request = model='gpt-4o' messages=[ChatMessage(role='user', content='Explain LLM distillation')] temperature=1.0 top_p=1.0 n=1 stream=True stop=None max_tokens=None presence_penalty=0.0 frequency_penalty=0.0 logit_bias=None user=None

    input = [
        Message(role=chat_message.role, content=chat_message.content)
        for chat_message in request.messages
    ]

    response_id = (
        f"chatcmpl-{int(time() * 1000)}"  # generate a pseudo-unique ID
    )
    timestamp = int(time())

    settings = GenerationSettings(
        temperature=request.temperature,
        max_new_tokens=request.max_tokens,
        stop_strings=request.stop,
        top_p=request.top_p,
        # num_return_sequences=request.n
    )

    response = await orchestrator(input, settings=settings)

    # Streaming through SSE (server-sent events with text/event-stream)
    if request.stream:

        async def generate_chunks():
            async for token in response:
                # OpenAI stream delta chunk
                choice = ChatCompletionChunkChoice(
                    delta=ChatCompletionChunkChoiceDelta(content=token)
                )
                chunk = ChatCompletionChunk(
                    id=response_id,
                    created=timestamp,
                    model=request.model,
                    choices=[choice],
                )
                yield f"data: {chunk.json()}\n\n"  # SSE data event
            yield "data: [DONE]\n\n"  # end of stream

        return StreamingResponse(
            generate_chunks(), media_type="text/event-stream"
        )

    # Non streaming
    message = ChatMessage(
        role=str(MessageRole.ASSISTANT), content=await response.to_str()
    )
    usage = ChatCompletionUsage()
    response = ChatCompletionResponse(
        id=response_id,
        created=timestamp,
        model=request.model,
        choices=[ChatCompletionChoice(message=message, finish_reason="stop")],
        usage=usage,
    )
    return response



from abc import ABC, abstractmethod
from ..model.entities import EngineMessage
from dataclasses import dataclass
from threading import Lock
from typing import Generic, Optional, TypeVar
from ..compat import override
from uuid import UUID

T = TypeVar("T")


@dataclass(frozen=True)
class MemoryChunk(Generic[T]):
    repository_key: str
    key: Optional[str]
    data: T


class MemoryStore(ABC, Generic[T]):
    @abstractmethod
    async def append(self, agent_id: UUID, data: T) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def reset(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def search(self, query: str) -> Optional[list[T]]:
        raise NotImplementedError()


class MessageMemory(MemoryStore[EngineMessage], ABC):
    def search(self, query: str) -> Optional[list[EngineMessage]]:
        raise NotImplementedError()


class RecentMessageMemory(MessageMemory):
    _lock: Lock
    _data: list[EngineMessage]

    def __init__(self, **kwargs) -> None:
        self._lock = Lock()
        self.reset()
        super().__init__(**kwargs)

    @override
    def append(self, data: EngineMessage) -> None:
        with self._lock:
            self._data.append(data)

    def reset(self) -> None:
        with self._lock:
            self._data = []

    @property
    def size(self) -> int:
        return len(self._data)

    @property
    def is_empty(self) -> bool:
        return not bool(self._data)

    @property
    def data(self) -> list[EngineMessage]:
        return self._data



from ...model.nlp.sentence import SentenceTransformerModel
from dataclasses import dataclass
from logging import Logger
from numpy import ndarray
from re import split
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from typing import Union


@dataclass(frozen=True, kw_only=True)
class TextPartition:
    data: str
    total_tokens: int
    embeddings: ndarray


class TextPartitioner:
    _PARAGRAPH_PATTERN = r"(?:\r\n|\r|\n){2,}"
    _model: SentenceTransformerModel
    _tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
    _logger: Logger
    _max_tokens: int
    _window_size: int
    _overlap_size: int

    def __init__(
        self,
        model: SentenceTransformerModel,
        logger: Logger,
        *args,
        max_tokens: int,
        overlap_size: int,
        window_size: int,
    ):
        assert model and model.tokenizer
        self._model = model
        self._tokenizer = model.tokenizer
        self._logger = logger
        self.configure(
            max_tokens=max_tokens,
            overlap_size=overlap_size,
            window_size=window_size,
        )

    def configure(
        self,
        *args,
        max_tokens: int,
        overlap_size: int,
        window_size: int,
    ) -> None:
        assert (
            max_tokens
            and window_size
            and overlap_size
            and max_tokens > 0
            and window_size > 0
            and overlap_size > 0
            and window_size < max_tokens
            and overlap_size < max_tokens
            and window_size > overlap_size
        )
        self._max_tokens = max_tokens
        self._window_size = window_size
        self._overlap_size = overlap_size

    async def __call__(
        self,
        text: str,
    ) -> list[TextPartition]:
        assert text
        partitions: list[TextPartition] = []

        for paragraph in split(self._PARAGRAPH_PATTERN, text):
            paragraph = paragraph.strip()
            if not paragraph:
                continue
            token_ids = self._tokenizer.encode(
                paragraph,
                add_special_tokens=False,
            )
            length = len(token_ids)
            if length <= self._max_tokens:
                embeddings = await self._model(paragraph)
                partitions.append(
                    TextPartition(
                        data=paragraph,
                        embeddings=embeddings,
                        total_tokens=length,
                    )
                )
            else:
                step = self._window_size - self._overlap_size

                for start in range(0, length, step):
                    window_token_ids = token_ids[
                        start : start + self._window_size
                    ]
                    section_length = len(window_token_ids)
                    section = self._tokenizer.decode(
                        window_token_ids, skip_special_tokens=True
                    )
                    embeddings = await self._model(section)
                    partitions.append(
                        TextPartition(
                            data=section,
                            embeddings=embeddings,
                            total_tokens=section_length,
                        )
                    )

        return partitions



from argparse import Namespace
from ...agent.loader import OrchestrationLoader
from ...cli import get_input
from ...cli.commands.model import token_generation
from ...event import EventStats
from ...memory.permanent import VectorFunction
from ...model import TextGenerationResponse
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.nlp.text.vendor import TextGenerationVendorModel
from ...server import agents_server
from contextlib import AsyncExitStack
from logging import Logger
from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.syntax import Syntax
from rich.theme import Theme
from typing import Optional
from uuid import uuid4
from jinja2 import Environment, FileSystemLoader
from os.path import dirname, join


async def agent_message_search(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    refresh_per_second: int,
) -> None:
    assert args.specifications_file
    _, _i = theme._, theme.icons

    specs_path = args.specifications_file
    agent_id = args.id
    participant_id = args.participant
    session_id = args.session

    assert agent_id and participant_id and session_id

    input_string = get_input(
        console,
        _i["user_input"] + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    if not input_string:
        return

    limit = args.limit

    async with AsyncExitStack() as stack:
        with console.status(
            _("Loading agent..."),
            spinner=theme.get_spinner("agent_loading"),
            refresh_per_second=refresh_per_second,
        ):
            logger.debug(
                f"Loading agent from {specs_path} for "
                f"participant {participant_id}"
            )

            orchestrator = await OrchestrationLoader.from_file(
                specs_path,
                agent_id=agent_id,
                hub=hub,
                logger=logger,
                participant_id=participant_id,
                stack=stack,
            )
            orchestrator = await stack.enter_async_context(orchestrator)

            assert orchestrator.engine_agent and orchestrator.engine.model_id

            can_access = args.skip_hub_access_check or hub.can_access(
                orchestrator.engine.model_id
            )
            is_local = not isinstance(
                orchestrator.engine, TextGenerationVendorModel
            )
            models = [
                hub.model(model_id) if is_local else model_id
                for model_id in orchestrator.model_ids
            ]

            console.print(
                theme.agent(orchestrator, models=models, can_access=can_access)
            )

            messages = await orchestrator.memory.search_messages(
                search=input_string,
                agent_id=agent_id,
                session_id=session_id,
                participant_id=participant_id,
                function=VectorFunction.L2_DISTANCE,
                limit=limit,
            )
            console.print(
                theme.search_message_matches(
                    participant_id, orchestrator, messages
                )
            )


async def agent_run(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    refresh_per_second: int,
) -> None:
    assert args.specifications_file
    _, _i = theme._, theme.icons

    specs_path = args.specifications_file
    use_async_generator = not args.use_sync_generator
    display_tokens = args.display_tokens or 0
    dtokens_pick = 10 if display_tokens > 0 else 0
    with_stats = args.stats and not args.quiet
    agent_id = args.id
    participant_id = args.participant
    session_id = args.session if not args.no_session else None
    load_recent_messages = (
        not args.skip_load_recent_messages and not args.no_session
    )
    load_recent_messages_limit = args.load_recent_messages_limit

    event_stats = EventStats()

    async def _event_listener(event):
        nonlocal event_stats
        event_stats.total_triggers += 1
        if event.type not in event_stats.triggers:
            event_stats.triggers[event.type] = 1
        else:
            event_stats.triggers[event.type] += 1

    async with AsyncExitStack() as stack:
        with console.status(
            _("Loading agent..."),
            spinner=theme.get_spinner("agent_loading"),
            refresh_per_second=refresh_per_second,
        ):
            logger.debug(
                f"Loading agent from {specs_path} for "
                f"participant {participant_id}"
            )

            orchestrator = await OrchestrationLoader.from_file(
                specs_path,
                agent_id=agent_id,
                hub=hub,
                logger=logger,
                participant_id=participant_id,
                stack=stack,
                disable_memory=args.no_session,
            )
            orchestrator.event_manager.add_listener(_event_listener)

            orchestrator = await stack.enter_async_context(orchestrator)

            logger.debug(
                f"Agent loaded from {specs_path}, models "
                f"used: {orchestrator.model_ids}, with recent message memory: "
                f"{'yes' if orchestrator.memory.has_recent_message else 'no'},"
                " with permanent message memory: yes, with session #"
                + str(orchestrator.memory.permanent_message.session_id)
                if orchestrator.memory.has_permanent_message
                else "no"
            )

        if not args.quiet:
            assert orchestrator.engine_agent and orchestrator.engine.model_id

            is_local = not isinstance(
                orchestrator.engine, TextGenerationVendorModel
            )

            can_access = (
                args.skip_hub_access_check
                or not is_local
                or hub.can_access(orchestrator.engine.model_id)
            )
            models = [
                hub.model(model_id) if is_local else model_id
                for model_id in orchestrator.model_ids
            ]

            console.print(
                theme.agent(orchestrator, models=models, can_access=can_access)
            )

        if not args.no_session:
            if session_id:
                await orchestrator.memory.continue_session(
                    session_id=session_id,
                    load_recent_messages=load_recent_messages,
                    load_recent_messages_limit=load_recent_messages_limit,
                )
            else:
                await orchestrator.memory.start_session()

        if (
            load_recent_messages
            and orchestrator.memory.has_recent_message
            and not orchestrator.memory.recent_message.is_empty
            and not args.quiet
        ):
            console.print(
                theme.recent_messages(
                    participant_id,
                    orchestrator,
                    orchestrator.memory.recent_message.data,
                )
            )

        input_string: Optional[str] = None
        in_conversation = False
        while not input_string or in_conversation:
            logger.debug(
                "Waiting for new message to add to orchestrator's existing "
                + str(orchestrator.memory.recent_message.size)
                if orchestrator.memory
                and orchestrator.memory.has_recent_message
                else "0" + " messages"
            )
            input_string = get_input(
                console,
                _i["user_input"] + " ",
                echo_stdin=not args.no_repl,
                force_prompt=in_conversation,
                is_quiet=args.quiet,
                tty_path=args.tty,
            )
            if not input_string:
                logger.debug("Finishing session with orchestrator")
                return

            logger.debug(f'Agent about to process input "{input_string}"')
            output = await orchestrator(
                input_string, use_async_generator=use_async_generator
            )

            if not args.quiet and not args.stats:
                console.print(_i["agent_output"] + " ", end="")

            if isinstance(output, TextGenerationResponse):
                if not use_async_generator:
                    console.print(await output.to_str())
                else:
                    await token_generation(
                        args=args,
                        console=console,
                        theme=theme,
                        logger=logger,
                        orchestrator=orchestrator,
                        event_stats=event_stats,
                        lm=orchestrator.engine,
                        input_string=input_string,
                        response=output,
                        dtokens_pick=dtokens_pick,
                        display_tokens=display_tokens,
                        with_stats=with_stats,
                    )
            else:
                console.print(output)

            if args.conversation:
                console.print("")
                if not in_conversation:
                    in_conversation = True
            else:
                break


async def agent_serve(
    args: Namespace,
    hub: HuggingfaceHub,
    logger: Logger,
    name: str,
    version: str,
) -> None:
    assert args.host and args.port and args.specifications_file
    specs_path = args.specifications_file

    async with AsyncExitStack() as stack:
        logger.debug(f"Loading agent from {specs_path}")

        orchestrator = await OrchestrationLoader.from_file(
            specs_path,
            agent_id=uuid4(),
            hub=hub,
            logger=logger,
            participant_id=uuid4(),
            stack=stack,
        )
        orchestrator = await stack.enter_async_context(orchestrator)

        logger.debug(f"Agent loaded from {specs_path}")
        server = agents_server(
            name=name,
            version=version,
            prefix_openai=args.prefix_openai,
            prefix_mcp=args.prefix_mcp,
            agents=[orchestrator],
            host=args.host,
            port=args.port,
            reload=args.reload,
            logger=logger,
        )
        await server.serve()


async def agent_init(args: Namespace, console: Console, theme: Theme) -> None:
    _ = theme._

    name = args.name or Prompt.ask(_("Agent name"))
    role = args.role or get_input(
        console,
        _("Agent role") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    if not role:
        return

    task = args.task or get_input(
        console,
        _("Agent task") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    instructions = args.instructions or get_input(
        console,
        _("Agent instructions") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )

    memory_recent = (
        args.memory_recent
        if args.memory_recent is not None
        else Confirm.ask(_("Use recent message memory?"))
    )
    memory_permanent = (
        args.memory_permanent
        if args.memory_permanent is not None
        else Prompt.ask(_("Permanent memory DSN"), default="")
    )
    memory_engine_model_id = (
        args.memory_engine_model_id
        or OrchestrationLoader.DEFAULT_SENTENCE_MODEL_ID
    )
    engine_uri = args.engine_uri or Prompt.ask(
        _("Engine URI"),
        default="microsoft/Phi-4-mini-instruct",
    )
    run_use_cache = (
        args.use_cache
        if args.use_cache is not None
        else Confirm.ask(_("Cache model locally?"))
    )
    run_skip_special_tokens = args.skip_special_tokens
    max_new_tokens = args.max_new_tokens or 1024

    data = dict(
        name=name,
        role=role,
        task=task,
        instructions=instructions,
        memory_recent=memory_recent,
        memory_permanent=memory_permanent,
        memory_engine_model_id=memory_engine_model_id,
        engine_uri=engine_uri,
        run_use_cache=run_use_cache,
        run_skip_special_tokens=run_skip_special_tokens,
        max_new_tokens=max_new_tokens,
    )

    env = Environment(
        loader=FileSystemLoader(join(dirname(__file__), "..", "..", "agent")),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = env.get_template("blueprint.toml")
    rendered = template.render(**data)
    console.print(Syntax(rendered, "toml"))



from ..event import Event, EventType
from collections import defaultdict
from inspect import iscoroutine
from typing import Awaitable, Callable, Iterable

Listener = Callable[[Event], Awaitable[None] | None]


class EventManager:
    _listeners: dict[EventType, list[Listener]]

    def __init__(self) -> None:
        self._listeners = defaultdict(list)

    def add_listener(
        self, listener: Listener, event_types: Iterable[EventType] | None = None
    ) -> None:
        types = list(event_types) if event_types else list(EventType)
        for event_type in types:
            self._listeners[event_type].append(listener)

    async def trigger(self, event: Event) -> None:
        for listener in self._listeners.get(event.type, []):
            result = listener(event)
            if iscoroutine(result):
                await result



from sympy import sympify


class CalculatorTool:
    """Safely evaluate arithmetic expressions using sympy."""

    def __init__(self) -> None:
        self.__name__ = "calculator"

    def __call__(self, expression: str) -> str:
        """Return the result of the expression as a string."""
        result = sympify(expression, evaluate=True)
        return str(result)


tool = CalculatorTool()


def calculator(expression: str) -> str:
    """
    Calculate the result of the arithmetic expression.

    Args:
        expression: Expression to calculate.

    Returns:
        Result of the calculated expression
    """
    return tool(expression)



from .....model import TextGenerationVendor
from .openai import OpenAIClient, OpenAIModel
from transformers import PreTrainedModel


class AnyScaleClient(OpenAIClient):
    def __init__(self, api_key: str, base_url: str | None = None):
        super().__init__(
            api_key=api_key,
            base_url=base_url or "https://api.endpoints.anyscale.com/v1",
        )


class AnyScaleModel(OpenAIModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        assert self._settings.access_token
        return AnyScaleClient(
            base_url=self._settings.base_url,
            api_key=self._settings.access_token,
        )



from asyncio import to_thread
from ....compat import override
from ....model import TextGenerationVendorStream
from ....model.entities import (
    GenerationSettings,
    Input,
    TransformerEngineSettings,
)
from ....model.nlp.text.generation import TextGenerationModel
from ....tool.manager import ToolManager
from dataclasses import replace
from logging import Logger
from typing import AsyncGenerator

try:
    from vllm import LLM, SamplingParams
except Exception:  # pragma: no cover - vllm may not be installed
    LLM = None
    SamplingParams = None


class VllmStream(TextGenerationVendorStream):
    def __init__(self, generator):
        super().__init__(generator)
        self._iterator = generator

    async def __anext__(self) -> str:
        try:
            chunk = await to_thread(next, self._iterator)
        except StopIteration:
            raise StopAsyncIteration
        return chunk


class VllmModel(TextGenerationModel):
    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(model_id, settings, logger)

    @property
    def supports_sample_generation(self) -> bool:
        return False

    def _load_model(self):
        assert LLM, "vLLM is not available"
        return LLM(
            model=self._model_id,
            tokenizer=self._settings.tokenizer_name_or_path or self._model_id,
            trust_remote_code=self._settings.trust_remote_code,
        )

    def _build_sampling_params(
        self, settings: GenerationSettings
    ) -> SamplingParams:
        assert SamplingParams, "vLLM is not available"
        return SamplingParams(
            temperature=settings.temperature,
            top_p=settings.top_p,
            top_k=settings.top_k,
            max_tokens=settings.max_new_tokens,
            stop=settings.stop_strings,
        )

    def _prompt(
        self, input: Input, system_prompt: str | None, tool: ToolManager | None
    ) -> str:
        inputs = super()._tokenize_input(
            input,
            system_prompt,
            context=None,
            tensor_format="pt",
            tool=tool,
        )
        return self._tokenizer.decode(
            inputs["input_ids"][0], skip_special_tokens=False
        )

    async def _stream_generator(
        self,
        prompt: str,
        settings: GenerationSettings,
    ) -> AsyncGenerator[str, None]:
        params = self._build_sampling_params(settings)
        iterator = self._model.generate([prompt], params, stream=True)
        stream = VllmStream(iter(iterator))
        async for chunk in stream:
            yield chunk

    def _string_output(
        self,
        prompt: str,
        settings: GenerationSettings,
    ) -> str:
        params = self._build_sampling_params(settings)
        results = list(self._model.generate([prompt], params))
        return results[0].outputs[0].text if results else ""

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
    ) -> TextGenerationVendorStream | str:
        settings = settings or GenerationSettings()
        prompt = self._prompt(input, system_prompt, tool)
        generation_settings = replace(settings, do_sample=False)
        if settings.use_async_generator:
            return await self._stream_generator(prompt, generation_settings)
        return self._string_output(prompt, generation_settings)



from ..agent import Role, Specification
from ..agent.engine import EngineAgent
from ..memory.manager import MemoryManager
from ..model.engine import Engine
from ..tool.manager import ToolManager
from ..event.manager import EventManager
from jinja2 import (
    Environment as TemplateEnvironment,
    FileSystemLoader,
    Template,
)
from os import linesep
from os.path import dirname, join
from typing import Any, Optional
from uuid import UUID


class Renderer:
    _TEMPLATES_DIR = "templates"
    _environment: TemplateEnvironment
    _clean_spaces: bool
    _templates: dict[str, Template] = {}

    def __init__(
        self, templates_path: Optional[str] = None, clean_spaces: bool = True
    ):
        self._clean_spaces = clean_spaces
        self._environment = TemplateEnvironment(
            loader=FileSystemLoader(
                templates_path or join(dirname(__file__), self._TEMPLATES_DIR)
            ),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def __call__(self, template_id: str, **kwargs) -> str:
        if template_id not in self._templates:
            self._templates[template_id] = self._environment.get_template(
                template_id
            )

        output = self._templates[template_id].render(**kwargs)

        if self._clean_spaces:
            output = linesep.join(line.strip() for line in output.splitlines())
        return output

    def from_string(
        self,
        template: str,
        template_vars: Optional[dict] = None,
        encoding: str = "utf-8",
    ) -> str:
        return (
            Template(template).render(**template_vars).encode(encoding)
            if template_vars
            else template
        )


class TemplateEngineAgent(EngineAgent):
    _renderer: Renderer

    def __init__(
        self,
        model: Engine,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        renderer: Renderer,
        *args,
        name: Optional[str] = None,
        id: Optional[UUID] = None,
    ):
        super().__init__(
            model,
            memory,
            tool,
            event_manager,
            name=name,
            id=id,
        )
        self._renderer = renderer

    def _prepare_call(
        self, specification: Specification, input: str, **kwargs: Any
    ) -> Any:
        template_id = specification.template_id or "agent.md"
        template_vars = specification.template_vars or {}
        template_vars.setdefault("name", self._name)
        template_vars.setdefault(
            "roles",
            (
                [
                    self._renderer.from_string(persona, template_vars)
                    for persona in specification.role.persona
                ]
                if isinstance(specification.role, Role)
                else [specification.role]
                if isinstance(specification.role, str)
                else specification.role
            )
            if template_vars
            else specification.role,
        )
        template_vars.setdefault(
            "task",
            self._renderer.from_string(specification.goal.task, template_vars)
            if specification.goal and template_vars
            else specification.goal.task
            if specification.goal
            else None,
        )
        template_vars.setdefault(
            "instructions",
            [
                self._renderer.from_string(instruction, template_vars)
                for instruction in specification.goal.instructions
            ]
            if specification.goal and template_vars
            else specification.goal.instructions
            if specification.goal
            else None,
        )
        template_vars.setdefault(
            "rules",
            [
                self._renderer.from_string(rule, template_vars)
                for rule in specification.rules
            ]
            if template_vars and specification.rules
            else specification.rules,
        )
        system_prompt = self._renderer(template_id, **template_vars)

        kwargs.setdefault("settings", specification.settings)
        kwargs.setdefault("system_prompt", system_prompt)
        return kwargs



from ....memory.partitioner.text import TextPartition
from ....memory.permanent import (
    PermanentMessage,
    PermanentMessageMemory,
    PermanentMessagePartition,
    PermanentMessageScored,
    Session,
    VectorFunction,
)
from ....memory.permanent.pgsql import PgsqlMemory
from ....model.entities import EngineMessage, EngineMessageScored
from datetime import datetime, timezone
from pgvector.psycopg import Vector
from typing import Optional
from uuid import UUID, uuid4


class PgsqlMessageMemory(PgsqlMemory[PermanentMessage], PermanentMessageMemory):
    @classmethod
    async def create_instance(
        cls,
        dsn: str,
        *args,
        pool_minimum: int = 1,
        pool_maximum: int = 10,
        pool_open: bool = True,
        **kwargs,
    ):
        memory = cls(
            dsn=dsn,
            composite_types=["message_author_type"],
            pool_minimum=pool_minimum,
            pool_maximum=pool_maximum,
            **kwargs,
        )
        if pool_open:
            await memory.open()
        return memory

    async def create_session(
        self, *args, agent_id: UUID, participant_id: UUID
    ) -> UUID:
        now_utc = datetime.now(timezone.utc)
        session = Session(
            id=uuid4(),
            agent_id=agent_id,
            participant_id=participant_id,
            messages=0,
            created_at=now_utc,
        )
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await cursor.execute(
                    """
                    INSERT INTO "sessions"(
                        "id",
                        "agent_id",
                        "participant_id",
                        "messages",
                        "created_at"
                    ) VALUES (
                        %s, %s, %s, %s, %s
                    )
                """,
                    (
                        str(session.id),
                        str(session.agent_id),
                        str(session.participant_id),
                        session.messages,
                        session.created_at,
                    ),
                )
                await cursor.close()
        return session.id

    async def continue_session_and_get_id(
        self,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        session_id: UUID,
    ) -> UUID:
        session_id = await self._fetch_field(
            "id",
            """
            SELECT "sessions"."id"
            FROM "sessions"
            WHERE "agent_id" = %s
            AND "participant_id" = %s
            AND "id" = %s
            LIMIT 1
        """,
            (str(agent_id), str(participant_id), str(session_id)),
        )
        assert session_id
        return session_id if isinstance(session_id, UUID) else UUID(session_id)

    async def append_with_partitions(
        self,
        engine_message: EngineMessage,
        *args,
        partitions: list[TextPartition],
    ) -> None:
        assert engine_message and partitions
        now_utc = datetime.now(timezone.utc)
        message = PermanentMessage(
            id=uuid4(),
            agent_id=engine_message.agent_id,
            model_id=engine_message.model_id,
            session_id=self._session_id,
            author=engine_message.message.role,
            data=engine_message.message.content,
            partitions=len(partitions),
            created_at=now_utc,
        )
        message_partitions = [
            PermanentMessagePartition(
                agent_id=message.agent_id,
                session_id=message.session_id,
                message_id=message.id,
                partition=i,
                data=p.data,
                embedding=p.embeddings,
                created_at=now_utc,
            )
            for i, p in enumerate(partitions)
        ]

        async with self._database.connection() as connection:
            async with connection.transaction():
                async with connection.cursor() as cursor:
                    await cursor.execute(
                        """
                        INSERT INTO "messages"(
                            "id",
                            "agent_id",
                            "model_id",
                            "session_id",
                            "author",
                            "data",
                            "partitions",
                            "created_at"
                        ) VALUES (
                            %s, %s, %s, %s, %s, %s, %s, %s
                        )
                    """,
                        (
                            str(message.id),
                            str(message.agent_id),
                            str(message.model_id),
                            str(message.session_id)
                            if message.session_id
                            else None,
                            str(message.author),
                            message.data,
                            message.partitions,
                            message.created_at,
                        ),
                    )

                    if message.session_id:
                        await cursor.execute(
                            """
                            UPDATE "sessions"
                            SET "messages" = "messages" + 1
                            WHERE "id" = %s
                        """,
                            (str(message.session_id),),
                        )

                    await cursor.executemany(
                        """
                        INSERT INTO "message_partitions"(
                            "agent_id",
                            "session_id",
                            "message_id",
                            "partition",
                            "data",
                            "embedding",
                            "created_at"
                        ) VALUES (
                            %s, %s, %s, %s, %s, %s, %s
                        )
                    """,
                        [
                            (
                                str(mp.agent_id),
                                str(mp.session_id) if mp.session_id else None,
                                str(mp.message_id),
                                mp.partition + 1,
                                mp.data,
                                Vector(mp.embedding),
                                mp.created_at,
                            )
                            for mp in message_partitions
                        ],
                    )

                    await cursor.close()

    async def get_recent_messages(
        self,
        session_id: UUID,
        participant_id: UUID,
        *args,
        limit: Optional[int] = None,
    ) -> list[EngineMessage]:
        messages = await self._fetch_all(
            PermanentMessage,
            """
            SELECT
                "messages"."id",
                "messages"."agent_id",
                "messages"."model_id",
                "messages"."session_id",
                "messages"."author",
                "messages"."data",
                "messages"."partitions",
                "messages"."created_at"
            FROM "sessions"
            INNER JOIN "messages" ON "sessions"."id" = "messages"."session_id"
            WHERE "sessions"."id" = %s
            AND "sessions"."participant_id" = %s
            AND "messages"."is_deleted" = FALSE
            ORDER BY "messages"."created_at" DESC
            LIMIT %s
        """,
            (str(session_id), str(participant_id), limit),
        )
        engine_messages = self._to_engine_messages(
            messages, limit=limit, reverse=True
        )
        return engine_messages

    async def search_messages(
        self,
        *args,
        search_partitions: list[TextPartition],
        agent_id: UUID,
        session_id: UUID,
        participant_id: UUID,
        function: VectorFunction,
        limit: Optional[int] = None,
    ) -> list[EngineMessageScored]:
        assert agent_id and session_id and participant_id and search_partitions
        search_function = str(function)
        search_vector = Vector(search_partitions[0].embeddings)
        messages = await self._fetch_all(
            PermanentMessageScored,
            f"""
            SELECT
                "messages"."id",
                "messages"."agent_id",
                "messages"."model_id",
                "messages"."session_id",
                "messages"."author",
                "messages"."data",
                "messages"."partitions",
                "messages"."created_at",
                {search_function}(
                    "message_partitions"."embedding",
                    %s
                ) AS "score"
            FROM "sessions"
            INNER JOIN "message_partitions" ON (
                "sessions"."id" = "message_partitions"."session_id"
            )
            INNER JOIN "messages" ON (
                "message_partitions"."message_id" = "messages"."id"
            )
            WHERE "sessions"."id" = %s
            AND "sessions"."participant_id" = %s
            AND "sessions"."agent_id" = %s
            AND "messages"."is_deleted" = FALSE
            ORDER BY "score" ASC
            LIMIT %s
        """,
            (
                search_vector,
                str(session_id),
                str(participant_id),
                str(agent_id),
                limit,
            ),
        )
        engine_messages = self._to_engine_messages(
            messages,
            limit=limit,
            scored=True,
        )
        return engine_messages



from .....compat import override
from .....model import TextGenerationVendor, TextGenerationVendorStream
from .....model.entities import (
    GenerationSettings,
    Message,
    Token,
    TokenDetail,
)
from .....model.nlp.text.vendor import TextGenerationVendorModel
from .....tool.manager import ToolManager
from openai import AsyncOpenAI, AsyncStream
from transformers import PreTrainedModel
from typing import AsyncIterator


class OpenAIStream(TextGenerationVendorStream):
    def __init__(self, stream: AsyncStream):
        super().__init__(stream.__aiter__())

    async def __anext__(self) -> Token | TokenDetail | str:
        chunk = await self._generator.__anext__()
        text = chunk.choices[0].delta.content or ""
        return text


class OpenAIClient(TextGenerationVendor):
    _client: AsyncOpenAI

    def __init__(self, api_key: str, base_url: str | None):
        self._client = AsyncOpenAI(base_url=base_url, api_key=api_key)

    @override
    async def __call__(
        self,
        model_id: str,
        messages: list[Message],
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
        use_async_generator: bool = True,
    ) -> AsyncIterator[Token | TokenDetail | str]:
        template_messages = self._template_messages(messages)
        client_stream = await self._client.chat.completions.create(
            model=model_id,
            messages=template_messages,
            stream=use_async_generator,
        )
        return OpenAIStream(stream=client_stream)


class OpenAIModel(TextGenerationVendorModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        assert self._settings.access_token
        return OpenAIClient(
            base_url=self._settings.base_url,
            api_key=self._settings.access_token,
        )



from ..model.entities import (
    AttentionImplementation,
    EngineUri,
    TextGenerationLoaderClass,
    TransformerEngineSettings,
    Vendor,
    WeightType,
)
from ..model.hubs.huggingface import HuggingfaceHub
from ..model.nlp.sentence import SentenceTransformerModel
from ..model.nlp.text.generation import TextGenerationModel
from ..secrets import KeyringSecrets
from contextlib import ContextDecorator, ExitStack
from logging import Logger
from typing import Any, get_args
from urllib.parse import urlparse, parse_qsl


class ModelManager(ContextDecorator):
    _hub: HuggingfaceHub
    _stack: ExitStack
    _logger: Logger
    _secrets: KeyringSecrets

    def __init__(
        self,
        hub: HuggingfaceHub,
        logger: Logger,
        secrets: KeyringSecrets | None = None,
    ):
        self._hub, self._logger = hub, logger
        self._stack = ExitStack()
        self._secrets = secrets or KeyringSecrets()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any | None,
    ):
        return self._stack.__exit__(exc_type, exc_value, traceback)

    def get_engine_settings(
        self,
        engine_uri,
        settings: dict | None = None,
        is_sentence_transformer: bool | None = None,
    ) -> TransformerEngineSettings:
        engine_settings_args = settings or {}

        if not is_sentence_transformer and not engine_uri.is_local:
            token = None
            if engine_uri.password and engine_uri.user:
                if engine_uri.user == "secret":
                    token = self._secrets.read(engine_uri.password)
                else:
                    token = None
            elif engine_uri.user:
                token = engine_uri.user

            if token:
                engine_settings_args.update(access_token=token)

        engine_settings = TransformerEngineSettings(**engine_settings_args)
        return engine_settings

    def load(
        self,
        engine_uri: EngineUri,
        *args,
        attention: AttentionImplementation | None = None,
        device: str | None = None,
        disable_loading_progress_bar: bool = False,
        is_sentence_transformer: bool | None = None,
        loader_class: TextGenerationLoaderClass | None = "auto",
        low_cpu_mem_usage: bool = False,
        quiet: bool = False,
        revision: str | None = None,
        special_tokens: list[str] | None = None,
        tokenizer: str | None = None,
        tokens: list[str] | None = None,
        trust_remote_code: bool | None = None,
        weight_type: WeightType = "auto",
    ) -> SentenceTransformerModel | TextGenerationModel:
        engine_settings_args = dict(
            cache_dir=self._hub.cache_dir,
            device=device,
            disable_loading_progress_bar=quiet or disable_loading_progress_bar,
            low_cpu_mem_usage=low_cpu_mem_usage,
            loader_class=loader_class,
            revision=revision,
            special_tokens=special_tokens or None,
            tokenizer_name_or_path=tokenizer,
            tokens=tokens or None,
            weight_type=weight_type,
        )
        if not is_sentence_transformer:
            engine_settings_args.update(
                attention=attention or None,
                trust_remote_code=trust_remote_code or None,
            )

        engine_settings = self.get_engine_settings(
            engine_uri,
            engine_settings_args,
            is_sentence_transformer=is_sentence_transformer,
        )
        return self.load_engine(
            engine_uri, engine_settings, is_sentence_transformer
        )

    def load_engine(
        self,
        engine_uri: EngineUri,
        engine_settings: TransformerEngineSettings,
        is_sentence_transformer: bool | None = None,
    ) -> SentenceTransformerModel | TextGenerationModel:
        model_load_args = dict(
            model_id=engine_uri.model_id,
            settings=engine_settings,
            logger=self._logger,
        )

        # Load local model, or lazy-import per vendor
        if is_sentence_transformer or engine_uri.is_local:
            model = (
                SentenceTransformerModel(**model_load_args)
                if is_sentence_transformer
                else TextGenerationModel(**model_load_args)
            )
        elif engine_uri.vendor == "openai":
            from ..model.nlp.text.vendor.openai import OpenAIModel

            model = OpenAIModel(**model_load_args)
        elif engine_uri.vendor == "openrouter":
            from ..model.nlp.text.vendor.openrouter import OpenRouterModel

            model = OpenRouterModel(**model_load_args)
        elif engine_uri.vendor == "anyscale":
            from ..model.nlp.text.vendor.anyscale import AnyScaleModel

            model = AnyScaleModel(**model_load_args)
        elif engine_uri.vendor == "together":
            from ..model.nlp.text.vendor.together import TogetherModel

            model = TogetherModel(**model_load_args)
        elif engine_uri.vendor == "deepseek":
            from ..model.nlp.text.vendor.deepseek import DeepSeekModel

            model = DeepSeekModel(**model_load_args)
        elif engine_uri.vendor == "deepinfra":
            from ..model.nlp.text.vendor.deepinfra import DeepInfraModel

            model = DeepInfraModel(**model_load_args)
        elif engine_uri.vendor == "groq":
            from ..model.nlp.text.vendor.groq import GroqModel

            model = GroqModel(**model_load_args)
        elif engine_uri.vendor == "ollama":
            from ..model.nlp.text.vendor.ollama import OllamaModel

            model = OllamaModel(**model_load_args)
        elif engine_uri.vendor == "huggingface":
            from ..model.nlp.text.vendor.huggingface import HuggingfaceModel

            model = HuggingfaceModel(**model_load_args)
        elif engine_uri.vendor == "hyperbolic":
            from ..model.nlp.text.vendor.hyperbolic import HyperbolicModel

            model = HyperbolicModel(**model_load_args)
        elif engine_uri.vendor == "litellm":
            from ..model.nlp.text.vendor.litellm import LiteLLMModel

            model = LiteLLMModel(**model_load_args)

        self._stack.enter_context(model)
        return model

    @staticmethod
    def parse_uri(uri: str) -> EngineUri:
        parsed = urlparse(uri)
        if not parsed.scheme:
            uri = f"ai://{uri}"
            parsed = urlparse(uri)

        if parsed.scheme != "ai":
            raise ValueError(f"Invalid scheme {parsed.scheme!r}, expected 'ai'")

        vendor = parsed.hostname
        if not vendor or vendor not in get_args(Vendor) or vendor == "local":
            vendor = None
        use_host = bool(vendor)
        path_prefixed = parsed.path.startswith("/")
        params = dict(parse_qsl(parsed.query))

        # urlparse() normalizes hostname to lowercase, so keep original case
        authority = parsed.netloc.rsplit("@", 1)[-1]
        hostname = authority.split(":", 1)[0]

        model_id = (
            hostname + ("/" if path_prefixed else "")
            if not vendor and hostname != "local"
            else ""
        ) + (parsed.path[1:] if path_prefixed else parsed.path)
        engine_uri = EngineUri(
            vendor=vendor,
            host=hostname if use_host else None,
            port=(parsed.port or None) if use_host else None,
            user=parsed.username or None,
            password=parsed.password or None,
            model_id=model_id,
            params=params,
        )
        return engine_uri



from argparse import Namespace
from ...cli import get_input
from ...model.entities import Token, TransformerEngineSettings
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.nlp.text.generation import TextGenerationModel
from logging import Logger
from rich.console import Console
from rich.theme import Theme
from typing import Optional


async def tokenize(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> Optional[list[Token]]:
    assert args.tokenizer

    _, _i, _n = theme._, theme.icons, theme._n

    tokenizer_name_or_path = args.tokenizer
    with TextGenerationModel(
        tokenizer_name_or_path,
        settings=TransformerEngineSettings(
            args.device,
            hub.cache_dir,
            tokenizer_name_or_path=tokenizer_name_or_path,
            tokens=args.token
            if args.token and isinstance(args.token, list)
            else None,
            special_tokens=args.special_token
            if args.special_token and isinstance(args.special_token, list)
            else None,
            auto_load_model=False,
            auto_load_tokenizer=True,
            disable_loading_progress_bar=args.disable_loading_progress_bar,
            low_cpu_mem_usage=args.low_cpu_mem_usage,
            loader_class=args.loader_class,
            weight_type=args.weight_type,
        ),
        logger=logger,
    ) as lm:
        logger.debug(f"Loaded tokenizer {lm.tokenizer_config.__repr__()}")
        console.print(theme.tokenizer_config(lm.tokenizer_config))

        if args.save:
            paths = lm.save_tokenizer(args.save)
            total_files = len(paths)
            console.print(theme.saved_tokenizer_files(args.save, total_files))
            return

        input_string = get_input(
            console,
            _i["user_input"] + " ",
            echo_stdin=not args.no_repl,
            is_quiet=args.quiet,
        )
        if input_string:
            logger.debug(f"Loaded model {lm.config.__repr__()}")
            tokens = lm.tokenize(input_string)

            panel = theme.tokenizer_tokens(
                tokens,
                lm.tokenizer_config.tokens,
                lm.tokenizer_config.special_tokens,
                display_details=True,
            )
            console.print(panel)



from ..model.entities import ToolCall, ToolCallResult, ToolFormat
from ..tool import ToolSet
from ..tool.calculator import calculator
from ..tool.parser import ToolCallParser
from types import FunctionType
from typing import Sequence, Tuple


class ToolManager:
    _parser: ToolCallParser
    _tools: dict[str, FunctionType] | None

    @classmethod
    def create_instance(
        cls,
        *args,
        eos_token: str | None = None,
        enable_tools: list[str] | None = None,
        tool_format: ToolFormat | None = None,
        available_toolsets: Sequence[ToolSet] | None = None,
    ):
        enabled_toolsets: list[ToolSet] | None = None

        if not available_toolsets:
            available_toolsets = [ToolSet(tools=[calculator])]

        if enable_tools:
            enabled_toolsets = []
            for toolset in available_toolsets:
                prefix = f"{toolset.namespace}." if toolset.namespace else ""
                tools = [
                    tool
                    for tool in toolset.tools
                    if f"{prefix}{tool.__name__}" in enable_tools
                ]
                if tools:
                    enabled_toolsets.append(
                        ToolSet(tools=tools, namespace=toolset.namespace)
                    )

        parser = ToolCallParser(eos_token=eos_token, tool_format=tool_format)
        return cls(
            parser=parser,
            toolsets=enabled_toolsets,
        )

    @property
    def is_empty(self) -> bool:
        return not bool(self._tools)

    @property
    def tools(self) -> list[FunctionType] | None:
        return list(self._tools.values()) if self._tools else None

    def __init__(
        self,
        *args,
        parser: ToolCallParser,
        toolsets: Sequence[ToolSet] | None = None,
    ):
        self._parser = parser
        self._tools = None

        if toolsets:
            self._tools = {}
            for toolset in toolsets:
                prefix = f"{toolset.namespace}." if toolset.namespace else ""
                for tool in toolset.tools:
                    self._tools[f"{prefix}{tool.__name__}"] = tool

    def set_eos_token(self, eos_token: str) -> None:
        self._parser.set_eos_token(eos_token)

    def __call__(
        self, text: str
    ) -> Tuple[list[ToolCall] | None, list[ToolCallResult] | None]:
        tool_calls = self._parser(text)
        if not tool_calls:
            return None, None

        tool_results: list[ToolCallResult] = []
        for tool_call in tool_calls:
            tool = self._tools.get(tool_call.name, None)
            if tool:
                result = (
                    tool(*tool_call.arguments.values())
                    if tool_call.arguments
                    else tool()
                )
                tool_results.append(
                    ToolCallResult(
                        call=tool_call,
                        name=tool_call.name,
                        arguments=tool_call.arguments,
                        result=result,
                    )
                )
        return tool_calls, tool_results



from ..model.entities import MessageRole
from pydantic import BaseModel, Field
from typing import Optional, Union


class ChatMessage(BaseModel):
    role: MessageRole
    content: str


class ChatCompletionRequest(BaseModel):
    model: str = Field(
        ..., description="ID of the model to use for generating the completion"
    )
    messages: list[ChatMessage] = Field(
        ..., description="List of messages in the conversation"
    )
    temperature: Optional[float] = Field(
        1.0, ge=0.0, le=2.0, description="Sampling temperature"
    )
    top_p: Optional[float] = Field(
        1.0, ge=0.0, le=1.0, description="Nucleus sampling probability"
    )
    n: Optional[int] = Field(
        1, ge=1, description="Number of completions to generate"
    )
    stream: Optional[bool] = Field(
        False, description="Whether to stream back partial progress"
    )
    stop: Optional[Union[str, list[str]]] = Field(
        None,
        description="Sequence where the API will stop generating further tokens",
    )
    max_tokens: Optional[int] = Field(
        None, ge=1, description="Maximum tokens to generate in the completion"
    )
    presence_penalty: Optional[float] = Field(
        0.0,
        ge=-2.0,
        le=2.0,
        description="Penalty for new tokens based on whether they appear in text so far",
    )
    frequency_penalty: Optional[float] = Field(
        0.0,
        ge=-2.0,
        le=2.0,
        description="Penalty for new tokens based on their frequency in text so far",
    )
    logit_bias: Optional[dict[str, int]] = Field(
        None,
        description="Modify the likelihood of specified tokens appearing in the completion",
    )
    user: Optional[str] = Field(
        None, description="Unique identifier representing your end-user"
    )


class ChatCompletionChunkChoiceDelta(BaseModel):
    content: str


class ChatCompletionChoice(BaseModel):
    index: int = 0
    message: ChatMessage
    finish_reason: str


class ChatCompletionChunkChoice(BaseModel):
    index: int = 0
    delta: ChatCompletionChunkChoiceDelta


class ChatCompletionChunk(BaseModel):
    id: str
    object: str = "chat.completion.chunk"
    created: int
    model: str
    choices: list[ChatCompletionChunkChoice]


class ChatCompletionUsage(BaseModel):
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0


class ChatCompletionResponse(BaseModel):
    id: str
    object: str = "chat.completion"
    created: int
    model: str
    choices: list[ChatCompletionChoice]
    usage: ChatCompletionUsage


class ModelInfo(BaseModel):
    id: str
    object: str = "model"
    created: int
    owned_by: str
    permission: list[dict]


class ModelList(BaseModel):
    object: str = "list"
    data: list[ModelInfo]



from ...compat import override
from ...model import TextGenerationVendor
from ...model.entities import (
    GenerationSettings,
    ImageEntity,
    ImageTextGenerationLoaderClass,
    Input,
    MessageRole,
)
from ...model.nlp import BaseNLPModel
from ...model.vision import BaseVisionModel
from ...model.transformer import TransformerModel
from PIL import Image
from torch import no_grad, Tensor
from transformers import (
    AutoImageProcessor,
    AutoModelForImageClassification,
    AutoModelForImageTextToText,
    AutoModelForVision2Seq,
    AutoProcessor,
    Gemma3ForConditionalGeneration,
    PreTrainedModel,
    Qwen2VLForConditionalGeneration,
    VisionEncoderDecoderModel as HFVisionEncoderDecoderModel,
)
from transformers.tokenization_utils_base import BatchEncoding
from typing import Literal


# model predicts one of the 1000 ImageNet classes
class ImageClassificationModel(BaseVisionModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_id,
            # default behavior in transformers v4.48
            use_fast=True,
        )
        model = AutoModelForImageClassification.from_pretrained(
            self._model_id,
            device_map=self._device,
        )
        return model

    @override
    async def __call__(
        self,
        image_source: str | Image.Image,
        tensor_format: Literal["pt"] = "pt",
    ) -> ImageEntity:
        image = BaseVisionModel._get_image(image_source)
        inputs = self._processor(image, return_tensors=tensor_format)

        with no_grad():
            logits = self._model(**inputs).logits

        label_index = logits.argmax(dim=1).item()
        return ImageEntity(label=self._model.config.id2label[label_index])


class ImageToTextModel(TransformerModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_id,
            # default behavior in transformers v4.48
            use_fast=True,
        )
        model = AutoModelForVision2Seq.from_pretrained(
            self._model_id,
            device_map=self._device,
        )
        return model

    def _tokenize_input(
        self,
        input: Input,
        context: str | None = None,
        tensor_format: Literal["pt"] = "pt",
        **kwargs,
    ) -> dict[str, Tensor] | BatchEncoding | Tensor:
        raise NotImplementedError()

    @override
    async def __call__(
        self,
        image_source: str | Image.Image,
        *,
        skip_special_tokens: bool = True,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        image = BaseVisionModel._get_image(image_source)
        inputs = self._processor(images=image, return_tensors=tensor_format)
        output_ids = self._model.generate(**inputs)
        caption = self._tokenizer.decode(
            output_ids[0], skip_special_tokens=skip_special_tokens
        )
        return caption


class ImageTextToTextModel(ImageToTextModel):
    _loaders: dict[ImageTextGenerationLoaderClass, type[PreTrainedModel]] = {
        "auto": AutoModelForImageTextToText,
        "qwen2": Qwen2VLForConditionalGeneration,
        "gemma3": Gemma3ForConditionalGeneration,
    }

    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        assert self._settings.loader_class in self._loaders, (
            f"Unrecognized loader {self._settings.loader_class}"
        )

        self._processor = AutoProcessor.from_pretrained(
            self._model_id,
            use_fast=True,
        )

        loader = self._loaders[self._settings.loader_class]
        model = loader.from_pretrained(
            self._model_id,
            torch_dtype=BaseNLPModel._get_weight_type(
                self._settings.weight_type
            ),
            device_map=self._device,
        )
        return model

    @override
    async def __call__(
        self,
        image_source: str | Image.Image,
        prompt: str,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        *,
        skip_special_tokens: bool = True,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        image = BaseVisionModel._get_image(image_source).convert("RGB")

        messages = []
        if system_prompt:
            messages.append(
                {
                    "role": str(MessageRole.SYSTEM),
                    "content": [{"type": "text", "text": system_prompt}],
                }
            )
        messages.append(
            {
                "role": str(MessageRole.USER),
                "content": [
                    {"type": "image", "image": image},
                    {"type": "text", "text": prompt},
                ],
            }
        )

        text = self._processor.apply_chat_template(
            messages, tokenize=False, add_generation_prompt=True
        )
        inputs = self._processor(
            text=[text],
            images=image,
            videos=None,
            padding=True,
            return_tensors=tensor_format,
        )
        inputs = inputs.to(self._device)
        generated_ids = self._model.generate(
            **inputs, max_new_tokens=settings.max_new_tokens
        )
        generated_ids_trimmed = [
            out_ids[len(in_ids) :]
            for in_ids, out_ids in zip(inputs.input_ids, generated_ids)
        ]
        output_text = self._processor.batch_decode(
            generated_ids_trimmed,
            skip_special_tokens=skip_special_tokens,
            clean_up_tokenization_spaces=False,
        )
        return output_text[0] if isinstance(output_text, list) else output_text


class VisionEncoderDecoderModel(ImageToTextModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_id,
            use_fast=True,
        )
        model = HFVisionEncoderDecoderModel.from_pretrained(
            self._model_id,
            device_map=self._device,
        )
        return model



from argparse import ArgumentParser, Namespace
from asyncio import run as run_in_loop
from asyncio.exceptions import CancelledError
from .. import license, name, site, version
from ..cli import CommandAbortException, has_input
from ..cli.commands.agent import (
    agent_message_search,
    agent_run,
    agent_serve,
    agent_init,
)
from ..cli.commands.cache import cache_delete, cache_download, cache_list
from ..cli.commands.memory import (
    memory_document_index,
    memory_embeddings,
    memory_search,
)
from ..cli.commands.model import (
    model_display,
    model_install,
    model_run,
    model_search,
    model_uninstall,
)
from ..cli.commands.tokenizer import tokenize
from ..cli.theme.fancy import FancyTheme
from ..memory.permanent import VectorFunction
from ..model.entities import (
    AttentionImplementation,
    DistanceType,
    TextGenerationLoaderClass,
    User,
    WeightType,
)
from ..model.hubs.huggingface import HuggingfaceHub
from ..model.transformer import TransformerModel
from ..utils import logger_replace
import gettext
from gettext import translation
from importlib.util import find_spec
from locale import getlocale
from logging import basicConfig, DEBUG, getLogger, INFO, Logger, WARNING
from os import getenv
from os.path import join
from pathlib import Path
from rich.console import Console
from rich.logging import RichHandler
from rich.prompt import Confirm, Prompt
from rich.theme import Theme
from typing import get_args, Optional
from uuid import uuid4
from warnings import filterwarnings


class CLI:
    _REFRESH_RATE = 4

    def __init__(self, logger: Logger):
        self._name = name()
        self._site = site()
        self._version = version()
        self._license = license()
        self._logger = logger

        cache_dir = HuggingfaceHub.DEFAULT_CACHE_DIR
        default_locale, _ = getlocale()
        default_locales_path = join(
            Path(__file__).resolve().parents[3], "locale"
        )
        default_device = TransformerModel.get_default_device()
        self._parser = CLI._create_parser(
            default_device, cache_dir, default_locales_path, default_locale
        )

    @staticmethod
    def _create_parser(
        default_device: str,
        cache_dir: str,
        default_locales_path: str,
        default_locale: str,
    ):
        global_parser = ArgumentParser(add_help=False)
        global_parser.add_argument(
            "--cache-dir",
            default=cache_dir,
            type=str,
            help=f"Path to huggingface cache hub (defaults to {cache_dir}, "
            "can also be specified with $HF_HUB_CACHE)",
        )
        global_parser.add_argument(
            "--device",
            type=str,
            required=False,
            default=default_device,
            help="Device to use (cpu, cuda, mps). Defaults to "
            + default_device,
        )
        global_parser.add_argument(
            "--disable-loading-progress-bar",
            action="store_true",
            default=False,
            help="If specified, the shard loading progress bar "
            "will not be shown",
        )
        global_parser.add_argument(
            "--hf-token",
            type=str,
            default=getenv("HF_ACCESS_TOKEN"),
            help="Your Huggingface access token",
        )
        global_parser.add_argument(
            "--locale",
            type=str,
            default=default_locale,
            help=f"Language to use (defaults to {default_locale})",
        )
        global_parser.add_argument(
            "--loader-class",
            type=str,
            default="auto",
            choices=get_args(TextGenerationLoaderClass),
            help='Loader class to use (defaults to "auto")',
        )
        global_parser.add_argument(
            "--locales",
            type=str,
            default=default_locales_path,
            help=f"Path to locale files (defaults to {default_locales_path})",
        )
        global_parser.add_argument(
            "--low-cpu-mem-usage",
            action="store_true",
            default=False,
            help="If specified, loads the model using ~1x model size "
            "CPU memory",
        )
        global_parser.add_argument(
            "--login",
            action="store_true",
            help="Login to main hub (huggingface)",
        )
        global_parser.add_argument(
            "--no-repl",
            action="store_true",
            help="Don't echo input coming from stdin",
        )
        global_parser.add_argument(
            "--quiet",
            "-q",
            default=False,
            action="store_true",
            help="If specified, no welcome screen and only model output is "
            "displayed in model run (sets "
            + ", ".join(
                [
                    "--disable-loading-progress-bar",
                    "--skip-hub-access-check",
                    "--skip-special-tokens",
                ]
            )
            + " automatically)",
        )
        global_parser.add_argument(
            "--revision",
            type=str,
            help="Model revision to use",
        )
        global_parser.add_argument(
            "--skip-hub-access-check",
            action="store_true",
            default=False,
            help="If specified, skip hub model access check",
        )
        global_parser.add_argument(
            "--verbose", "-v", action="count", help="Set verbosity"
        )
        global_parser.add_argument(
            "--weight-type",
            type=str,
            choices=get_args(WeightType),
            help="Weight type to use (defaults to best available)",
        )

        parser = ArgumentParser(
            description="Avalan CLI", parents=[global_parser]
        )

        command_parsers = parser.add_subparsers(dest="command")

        # Memory options shared by commands: memory embeddings, memory document
        memory_partitions_parser = ArgumentParser(add_help=False)
        memory_partitions_display_group = (
            memory_partitions_parser.add_mutually_exclusive_group()
        )
        memory_partitions_display_group.add_argument(
            "--no-display-partitions",
            action="store_true",
            default=False,
            help="If specified, don't display memory partitions",
        )
        memory_partitions_display_group.add_argument(
            "--display-partitions",
            default=6,
            type=int,
            help="Display up to this many partitions, if more summarize",
        )
        memory_partitions_parser.add_argument(
            "--partition",
            action="store_true",
            default=False,
            help="If specified, partition string",
        )
        memory_partitions_parser.add_argument(
            "--partition-max-tokens",
            default=500,
            type=int,
            help="Maximum number of tokens to allow on each partition",
        )
        memory_partitions_parser.add_argument(
            "--partition-overlap",
            default=125,
            type=int,
            help="How many tokens can potentially overlap in "
            "different partitions",
        )
        memory_partitions_parser.add_argument(
            "--partition-window",
            default=250,
            type=int,
            help="Number of tokens per window when partitioning",
        )

        # Model options shared by commands: memory embeddings, model
        model_options_parser = ArgumentParser(add_help=False)
        model_options_parser.add_argument(
            "model",
            type=str,
            help="Model to use",
        )
        model_options_parser.add_argument(
            "--load",
            action="store_true",
            help="If specified, load model and show more information",
        )
        model_options_parser.add_argument(
            "--special-token",
            type=str,
            action="append",
            help="Special token to add to tokenizer, only when model is loaded",
        )
        model_options_parser.add_argument(
            "--token",
            type=str,
            action="append",
            help="Token to add to tokenizer, only when model is loaded",
        )
        model_options_parser.add_argument(
            "--tokenizer",
            type=str,
            help="Path to tokenizer to use instead of model's default, only "
            "if model is loaded",
        )

        # Inference options shared by commands: agent run, model run
        model_inference_display_parser = ArgumentParser(add_help=False)
        model_inference_display_parser.add_argument(
            "--display-pause",
            type=int,
            nargs="?",
            const=500,  # 500 is the default if argument present but no value
            default=None,
            help="Pause (in ms.) when cycling through selected tokens as "
            "defined by --display-probabilities",
        )
        model_inference_display_parser.add_argument(
            "--display-probabilities",
            action="store_true",
            help="If --display-tokens specified, show also the token "
            "probability distribution",
        )
        model_inference_display_parser.add_argument(
            "--display-probabilities-maximum",
            type=float,
            default=0.8,
            help="When --display-probabilities is used, select tokens which "
            "logit probability is no higher than this value. "
            "Defaults to 0.8",
        )
        model_inference_display_parser.add_argument(
            "--display-probabilities-sample-minimum",
            type=float,
            default=0.1,
            help="When --display-probabilities is used, select tokens that "
            "have alternate tokens with a logit probability at least or "
            "higher than this value. Defaults to 0.1",
        )
        model_inference_display_parser.add_argument(
            "--display-time-to-n-token",
            type=int,
            nargs="?",
            const=256,  # 256 is the default if argument present but no value
            default=None,
            help="Display the time it takes to reach the given Nth token "
            "(defaults to 256)",
        )
        model_inference_display_parser.add_argument(
            "--display-tokens",
            type=int,
            nargs="?",
            const=15,  # 15 is the default if argument present but no value
            default=None,
            help="How many tokens with full information to display at a time",
        )

        # Agent command
        agent_parser = command_parsers.add_parser(
            name="agent",
            description="Manage AI agents",
            parents=[global_parser],
        )
        agent_command_parsers = agent_parser.add_subparsers(
            dest="agent_command"
        )

        agent_message_parser = agent_command_parsers.add_parser(
            name="message",
            description="Manage AI agent messages",
            parents=[global_parser],
        )
        agent_message_command_parsers = agent_message_parser.add_subparsers(
            dest="agent_message_command"
        )

        agent_message_search_parser = agent_message_command_parsers.add_parser(
            name="search",
            description="Search within an agent's message memory",
            parents=[global_parser],
        )
        agent_message_search_parser.add_argument(
            "specifications_file",
            type=str,
            help="File that holds the agent specifications",
        )
        agent_message_search_parser.add_argument(
            "--function",
            type=VectorFunction,
            choices=list(VectorFunction),
            required=True,
            default=VectorFunction.L2_DISTANCE,
            help="Vector function to use for searching",
        )
        agent_message_search_parser.add_argument(
            "--id", type=str, required=True
        )
        agent_message_search_parser.add_argument(
            "--limit",
            type=int,
            help="If specified, load up to these many recent messages",
        )
        agent_message_search_parser.add_argument(
            "--participant",
            type=str,
            required=True,
            help="Search messages with given participant",
        )
        agent_message_search_parser.add_argument(
            "--session",
            type=str,
            required=True,
            help="Search within the given session",
        )

        agent_run_parser = agent_command_parsers.add_parser(
            name="run",
            description="Run an AI agent",
            parents=[global_parser, model_inference_display_parser],
        )
        agent_run_parser.add_argument(
            "specifications_file",
            type=str,
            help="File that holds the agent specifications",
        )
        agent_run_parser.add_argument(
            "--conversation",
            action="store_true",
            default=False,
            help="Activate conversation mode with the agent",
        )
        agent_run_parser.add_argument(
            "--id", type=str, help="Use given ID as the agent ID"
        )
        agent_session_group = agent_run_parser.add_mutually_exclusive_group()
        agent_session_group.add_argument(
            "--no-session",
            action="store_true",
            default=False,
            help="If specified, don't use sessions in persistent "
            "message memory",
        )
        agent_session_group.add_argument(
            "--session",
            type=str,
            help="Continue the conversation on the given session",
        )

        agent_run_parser.add_argument(
            "--skip-load-recent-messages",
            default=False,
            action="store_true",
            help="If specified, skips loading recent messages",
        )
        agent_run_parser.add_argument(
            "--load-recent-messages-limit",
            type=int,
            help="If specified, load up to these many recent messages",
        )
        agent_run_parser.add_argument(
            "--participant",
            default=uuid4(),
            help="If specified, this is the participant ID interacting with "
            "the agent",
        )
        agent_run_parser.add_argument(
            "--stats",
            action="store_true",
            default=False,
            help="Show token generation statistics for agent output",
        )
        agent_run_parser.add_argument(
            "--sync",
            dest="use_sync_generator",
            action="store_true",
            default=False,
            help="Don't use an async generator (token streaming)",
        )
        agent_run_parser.add_argument(
            "--tty",
            default="/dev/tty",
            help="TTY stream (only applicable if combining --conversation "
            "with input piping)",
        )

        agent_serve_parser = agent_command_parsers.add_parser(
            name="serve",
            description="Serve an AI agent as an API endpoint",
            parents=[global_parser],
        )
        agent_serve_parser.add_argument(
            "specifications_file",
            type=str,
            help="File that holds the agent specifications",
        )
        agent_serve_parser.add_argument(
            "--host",
            default="127.0.0.1",
            type=str,
            help="Host (defaults to 127.0.0.1)",
        )
        agent_serve_parser.add_argument(
            "--port",
            default=9001,
            type=int,
            help="Port (defaults to 9001, HAL 9000+1)",
        )
        agent_serve_parser.add_argument(
            "--prefix-mcp",
            default="/mcp",
            type=str,
            help="URL prefix for MCP endpoints (defaults to /mcp)",
        )
        agent_serve_parser.add_argument(
            "--prefix-openai",
            default="/v1",
            type=str,
            help="URL prefix fir OpenAI endpoints (defaults to /v1)",
        )
        agent_serve_parser.add_argument(
            "--reload",
            action="store_true",
            default=False,
            help="Hot reload on code changes",
        )

        agent_init_parser = agent_command_parsers.add_parser(
            name="init",
            description="Create an agent definition",
            parents=[global_parser],
        )
        agent_init_parser.add_argument("--name", type=str, help="Agent name")
        agent_init_parser.add_argument("--role", type=str, help="Agent role")
        agent_init_parser.add_argument("--task", type=str, help="Agent task")
        agent_init_parser.add_argument(
            "--instructions", type=str, help="Agent instructions"
        )
        agent_init_parser.add_argument(
            "--memory-recent",
            dest="memory_recent",
            action="store_true",
            default=None,
            help="Enable recent message memory",
        )
        agent_init_parser.add_argument(
            "--no-memory-recent", dest="memory_recent", action="store_false"
        )
        agent_init_parser.add_argument(
            "--memory-permanent", type=str, help="Permanent memory DSN"
        )
        agent_init_parser.add_argument(
            "--memory-engine-model-id",
            type=str,
            help="Sentence transformer model for memory",
        )
        agent_init_parser.add_argument(
            "--engine-uri", type=str, help="Agent engine URI"
        )
        agent_init_parser.add_argument(
            "--use-cache",
            dest="use_cache",
            action="store_true",
            default=None,
            help="Cache model locally",
        )
        agent_init_parser.add_argument(
            "--no-cache", dest="use_cache", action="store_false"
        )
        agent_init_parser.add_argument(
            "--max-new-tokens", type=int, help="Max new tokens", default=None
        )
        agent_init_parser.add_argument(
            "--skip-special-tokens",
            action="store_true",
            help="Skip special tokens",
        )

        # Cache command
        cache_parser = command_parsers.add_parser(
            name="cache",
            description="Manage models cache",
            parents=[global_parser],
        )
        cache_command_parsers = cache_parser.add_subparsers(
            dest="cache_command"
        )
        cache_delete_parser = cache_command_parsers.add_parser(
            name="delete",
            description="Delete cached model data",
            parents=[global_parser],
        )
        cache_delete_parser.add_argument(
            "--delete",
            action="store_true",
            help="Actually delete. If not provided, a dry run is performed "
            "and data that would be deleted is shown, yet not deleted",
        )
        cache_delete_parser.add_argument(
            "--model",
            "-m",
            type=str,
            required=True,
            help="Model to delete",
        )
        cache_delete_parser.add_argument(
            "--delete-revision",
            type=str,
            action="append",
            help="Revision to delete",
        )
        cache_download_parser = cache_command_parsers.add_parser(
            name="download",
            description="Download model data to cache",
            parents=[global_parser],
        )
        cache_download_parser.add_argument(
            "--model",
            "-m",
            type=str,
            required=True,
            help="Model to load",
        )
        cache_list_parser = cache_command_parsers.add_parser(
            name="list",
            description="List cache contents",
            parents=[global_parser],
        )
        cache_list_parser.add_argument(
            "--model",
            type=str,
            action="append",
            help="Models to show content details on",
        )
        cache_list_parser.add_argument(
            "--summary",
            action="store_true",
            help="If specified, when showing one or more models show only "
            "summary",
        )

        # Deploy command
        deploy_parser = command_parsers.add_parser(
            name="deploy",
            description="Manage AI deployments",
            parents=[global_parser],
        )
        deploy_command_parsers = deploy_parser.add_subparsers(
            dest="deploy_command"
        )
        deploy_run_parser = deploy_command_parsers.add_parser(
            name="run",
            description="Perform a deployment",
            parents=[global_parser],
        )
        deploy_run_parser.add_argument(
            "deployment",
            type=str,
            help="Deployment to run",
        )

        # Flow command
        flow_parser = command_parsers.add_parser(
            name="flow", description="Manage AI flows", parents=[global_parser]
        )
        flow_command_parsers = flow_parser.add_subparsers(dest="flow_command")
        flow_run_parser = flow_command_parsers.add_parser(
            name="run", description="Run a given flow", parents=[global_parser]
        )
        flow_run_parser.add_argument(
            "flow",
            type=str,
            help="Flow to run",
        )

        # Memory command
        memory_parser = command_parsers.add_parser(
            name="memory", description="Manage memory", parents=[global_parser]
        )
        memory_command_parsers = memory_parser.add_subparsers(
            dest="memory_command"
        )
        memory_embeddings_parser = memory_command_parsers.add_parser(
            name="embeddings",
            description="Obtain and manipulate embeddings",
            parents=[
                global_parser,
                model_options_parser,
                memory_partitions_parser,
            ],
        )
        memory_embeddings_parser.add_argument(
            "--compare",
            type=str,
            action="append",
            help="If specified, compare embeddings with this string",
        )
        memory_embeddings_parser.add_argument(
            "--search",
            type=str,
            action="append",
            help="If specified, search across embeddings for this string",
        )
        memory_embeddings_parser.add_argument(
            "--search-k",
            default=1,
            type=int,
            help="How many nearest neighbors to obtain with search",
        )
        memory_embeddings_parser.add_argument(
            "--sort",
            type=DistanceType,
            choices=list(DistanceType),
            default=DistanceType.L2,
            help="Sort comparison results using the given similarity measure",
        )

        memory_search_parser = memory_command_parsers.add_parser(
            name="search",
            description="Search memories",
            parents=[
                global_parser,
                model_options_parser,
                memory_partitions_parser,
            ],
        )
        memory_search_parser.add_argument(
            "--dsn",
            type=str,
            required=True,
            help="PostgreSQL DSN for searching",
        )
        memory_search_parser.add_argument(
            "--participant",
            type=str,
            required=True,
            help="Participant ID to search",
        )
        memory_search_parser.add_argument(
            "--namespace", type=str, required=True, help="Namespace to search"
        )
        memory_search_parser.add_argument(
            "--function",
            type=VectorFunction,
            choices=list(VectorFunction),
            required=True,
            default=VectorFunction.L2_DISTANCE,
            help="Vector function to use for searching",
        )
        memory_search_parser.add_argument(
            "--limit", type=int, help="Return up to this many memories"
        )
        memory_doc_parser = memory_command_parsers.add_parser(
            name="document",
            description="Manage memory indexed documents",
        )
        memory_doc_command_parsers = memory_doc_parser.add_subparsers(
            dest="memory_document_command"
        )
        memory_doc_index_parser = memory_doc_command_parsers.add_parser(
            name="index",
            description="Add a document to the memory index",
            parents=[
                global_parser,
                model_options_parser,
                memory_partitions_parser,
            ],
        )
        memory_doc_index_parser.add_argument(
            "source",
            type=str,
            help="Source to index (an URL or a file path)",
        )
        memory_doc_index_parser.add_argument(
            "--partitioner",
            choices=["text", "code"],
            default="text",
            help="Partitioner to use when indexing a file",
        )
        memory_doc_index_parser.add_argument(
            "--language",
            type=str,
            help="Programming language for the code partitioner",
        )
        memory_doc_index_parser.add_argument(
            "--encoding",
            type=str,
            default="utf-8",
            help="File encoding used when reading a local file",
        )
        memory_doc_index_parser.add_argument(
            "--identifier",
            type=str,
            help="Identifier for the memory entry (defaults to the source)",
        )
        memory_doc_index_parser.add_argument(
            "--dsn",
            type=str,
            required=True,
            help="PostgreSQL DSN for storing the document",
        )
        memory_doc_index_parser.add_argument(
            "--participant",
            type=str,
            required=True,
            help="Participant ID for the memory entry",
        )
        memory_doc_index_parser.add_argument(
            "--namespace",
            type=str,
            required=True,
            help="Namespace for the memory entry",
        )

        # Model command
        model_parser = command_parsers.add_parser(
            name="model",
            description="Manage a model, showing details, loading or "
            "downloading it",
        )
        model_command_parsers = model_parser.add_subparsers(
            dest="model_command"
        )

        model_display_parser = model_command_parsers.add_parser(
            name="display",
            description="Show information about a model",
            parents=[global_parser, model_options_parser],
        )
        model_display_parser.add_argument(
            "--sentence-transformer",
            help="Load the model as a SentenceTransformer model",
            default=False,
            action="store_true",
        )
        model_display_parser.add_argument(
            "--summary",
            default=False,
            action="store_true",
        )

        model_command_parsers.add_parser(
            name="install",
            description="Install a model",
            parents=[global_parser, model_options_parser],
        )
        model_run_parser = model_command_parsers.add_parser(
            name="run",
            description="Run a model",
            parents=[
                global_parser,
                model_options_parser,
                model_inference_display_parser,
            ],
        )
        model_run_parser.add_argument(
            "--attention",
            type=str,
            choices=get_args(AttentionImplementation),
            help="Attention implementation to use (defaults to best available)",
        )
        model_run_parser.add_argument(
            "--do-sample",
            default=True,
            action="store_true",
            help="Tell if the token generation process should be "
            "deterministic or stochastic. When enabled, it's stochastic "
            "and uses probability distribution.",
        )
        model_run_parser.add_argument(
            "--enable-gradient-calculation",
            default=False,
            action="store_true",
            help="Enable gradient calculation.",
        )
        model_run_parser.add_argument(
            "--use-cache",
            default=False,
            action="store_true",
            help="Past key values are used to speed up decoding if applicable "
            "to model.",
        )
        model_run_parser.add_argument(
            "--max-new-tokens",
            default=10 * 1024,
            type=int,
            help="Maximum number of tokens to generate",
        )
        model_run_parser.add_argument(
            "--min-p",
            type=float,
            help="Minimum token probability, which will be scaled by the "
            "probability of the most likely token [0, 1]",
        )
        model_run_parser.add_argument(
            "--repetition-penalty",
            default=1.0,
            type=float,
            help="Exponential penalty on sequences not in the original input. "
            "Defaults to 1.0, which means no penalty.",
        )
        model_run_parser.add_argument(
            "--skip-special-tokens",
            default=False,
            action="store_true",
            help="If specified, skip special tokens when decoding",
        )
        model_run_parser.add_argument(
            "--system",
            type=str,
            help="Use this as system prompt",
        )
        model_run_parser.add_argument(
            "--start-thinking",
            default=False,
            action="store_true",
            help="If specified, assume model response starts with reasoning",
        )
        model_run_parser.add_argument(
            "--stop_on_keyword",
            type=str,
            action="append",
            help="Stop token generation when this keyword is found",
        )
        model_run_parser.add_argument(
            "--temperature",
            default=0.7,
            type=float,
            help="Temperature [0, 1]",
        )
        model_run_parser.add_argument(
            "--top-k",
            type=int,
            help="Number of highest probability vocabulary tokens to keep for "
            "top-k-filtering.",
        )
        model_run_parser.add_argument(
            "--top-p",
            type=float,
            help="If set to < 1, only the smallest set of most probable "
            "tokens with probabilities that add up to top_p or higher "
            "are kept for generation.",
        )
        model_run_parser.add_argument(
            "--trust-remote-code",
            action="store_true",
        )
        model_search_parser = model_command_parsers.add_parser(
            name="search",
            description="Search for models",
            parents=[global_parser],
        )
        model_search_parser.add_argument(
            "--search",
            type=str,
            required=False,
            help="Search for models matching given expression",
        )
        model_search_parser.add_argument(
            "--filter",
            type=str,
            help="Filter models on this (e.g: text-classification)",
        )
        model_search_parser.add_argument(
            "--limit",
            default=10,
            type=int,
            help="Maximum number of models to return",
        )
        model_uninstall_parser = model_command_parsers.add_parser(
            name="uninstall",
            description="Uninstall a model",
            parents=[global_parser, model_options_parser],
        )
        model_uninstall_parser.add_argument(
            "--delete",
            action="store_true",
            help="Actually delete. If not provided, a dry run is performed "
            "and data that would be deleted is shown, yet not deleted",
        )

        # Tokenizer command
        tokenizer_parser = command_parsers.add_parser(
            name="tokenizer",
            description="Manage tokenizers, loading, modifying and saving them",
            parents=[global_parser],
        )
        tokenizer_parser.add_argument(
            "--tokenizer",
            "-t",
            type=str,
            required=True,
            help="Tokenizer to load",
        )
        tokenizer_parser.add_argument(
            "--save",
            type=str,
            help="Save tokenizer (useful if modified via --special-token or "
            "--token) to given path, only if model is loaded",
        )
        tokenizer_parser.add_argument(
            "--special-token",
            type=str,
            action="append",
            help="Special token to add to tokenizer",
        )
        tokenizer_parser.add_argument(
            "--token",
            type=str,
            action="append",
            help="Token to add to tokenizer",
        )

        # Train command
        train_parser = command_parsers.add_parser(
            name="train", description="Training", parents=[global_parser]
        )
        train_command_parsers = train_parser.add_subparsers(
            dest="train_command"
        )
        train_run_parser = train_command_parsers.add_parser(
            name="run",
            description="Run a given training",
            parents=[global_parser],
        )
        train_run_parser.add_argument(
            "training",
            type=str,
            help="Training to run",
        )

        return parser

    async def __call__(self) -> None:
        args = self._parser.parse_args()

        current_locale, _ = getlocale()
        try:
            translator = translation(
                self._name, localedir=args.locales, languages=[args.locale]
            )
        except FileNotFoundError:
            translator = gettext

        assert self._logger is not None and isinstance(self._logger, Logger)
        theme = FancyTheme(translator.gettext, translator.ngettext)
        _ = theme._
        console = Console(theme=Theme(styles=theme.get_styles()))

        access_token = args.hf_token or Prompt.ask(theme.ask_access_token())
        assert access_token

        hub = HuggingfaceHub(access_token, args.cache_dir, self._logger)

        try:
            await self._main(args, theme, console, hub)
        except (CancelledError, KeyboardInterrupt, CommandAbortException):
            if not args.quiet:
                console.print(theme.bye())

    async def _main(
        self,
        args: Namespace,
        theme: Theme,
        console: Console,
        hub: HuggingfaceHub,
        suggest_login: bool = False,
    ) -> None:
        user: Optional[User] = None
        _ = theme._

        verbosity = args.verbose or 0
        log_level = (
            DEBUG if verbosity >= 2 else INFO if verbosity >= 1 else WARNING
        )
        previous_log_level = self._logger.getEffectiveLevel()

        self._logger.setLevel(log_level)

        if find_spec("sentence_transformers"):
            logger_replace(self._logger, ["sentence_transformers"])

        if find_spec("httpx"):
            logger_replace(self._logger, ["httpx"])

        filterwarnings(
            "ignore",
            message=r".*`do_sample` is set to `False`. "
            r"However, `temperature` is set.*",
        )

        suggest_login = suggest_login and not has_input(console)
        if args.login or (
            suggest_login
            and Confirm.ask(theme.ask_login_to_hub(), default=False)
        ):
            with console.status(
                theme.logging_in(hub.domain),
                spinner=(theme.get_spinner("connecting")),
                refresh_per_second=self._REFRESH_RATE,
            ):
                hub.login()
                user = hub.user()

        if not args.quiet:
            console.print(
                theme.welcome(
                    self._site.geturl(),
                    self._name,
                    self._version,
                    self._license,
                    user,
                )
            )

        match args.command:
            case "agent":
                subcommand = args.agent_command or "run"
                match subcommand:
                    case "message":
                        innercommand = args.agent_message_command or "search"
                        match innercommand:
                            case "search":
                                await agent_message_search(
                                    args,
                                    console,
                                    theme,
                                    hub,
                                    self._logger,
                                    refresh_per_second=self._REFRESH_RATE,
                                )
                    case "run":
                        await agent_run(
                            args,
                            console,
                            theme,
                            hub,
                            self._logger,
                            refresh_per_second=self._REFRESH_RATE,
                        )
                    case "serve":
                        await agent_serve(
                            args,
                            hub,
                            self._logger,
                            self._name,
                            str(self._version),
                        )
                    case "init":
                        await agent_init(args, console, theme)
            case "cache":
                subcommand = args.cache_command or "list"
                match subcommand:
                    case "delete":
                        cache_delete(args, console, theme, hub)
                    case "download":
                        cache_download(args, console, theme, hub)
                    case "list":
                        cache_list(args, console, theme, hub)
            case "memory":
                subcommand = args.memory_command or "embeddings"
                match subcommand:
                    case "document":
                        innercommand = args.memory_document_command or "index"
                        match innercommand:
                            case "index":
                                await memory_document_index(
                                    args, console, theme, hub, self._logger
                                )
                    case "search":
                        await memory_search(
                            args, console, theme, hub, self._logger
                        )
                    case "embeddings":
                        await memory_embeddings(
                            args, console, theme, hub, self._logger
                        )
            case "model":
                subcommand = args.model_command or "display"
                match subcommand:
                    case "display":
                        model_display(args, console, theme, hub, self._logger)
                    case "install":
                        model_install(args, console, theme, hub)
                    case "run":
                        await model_run(args, console, theme, hub, self._logger)
                    case "search":
                        await model_search(
                            args, console, theme, hub, self._REFRESH_RATE
                        )
                    case "uninstall":
                        model_uninstall(args, console, theme, hub)

            case "tokenizer":
                await tokenize(args, console, theme, hub, self._logger)

        self._logger.setLevel(previous_log_level)


def main() -> None:
    basicConfig(
        level=INFO,
        format="%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[RichHandler(rich_tracebacks=True)],
    )
    logger = getLogger(name())

    cli = CLI(logger)
    run_in_loop(cli())


if __name__ == "__main__":
    main()



from ...compat import override
from ...model import TextGenerationVendor
from ...model.entities import ImageEntity, EngineSettings
from ...model.vision import BaseVisionModel
from ...model.vision.image import ImageClassificationModel
from logging import Logger
from PIL import Image
from torch import tensor
from transformers import (
    AutoImageProcessor,
    AutoModelForObjectDetection,
    PreTrainedModel,
)
from typing import Literal


class ObjectDetectionModel(ImageClassificationModel):
    def __init__(
        self,
        model_id: str,
        settings: EngineSettings | None = None,
        revision: Literal["no_timm"] = "no_timm",
        logger: Logger | None = None,
    ):
        self._revision = revision
        super().__init__(model_id, settings, logger=logger)

    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_id,
            revision=self._revision,
            # default behavior in transformers v4.48
            use_fast=True,
        )
        model = AutoModelForObjectDetection.from_pretrained(
            self._model_id,
            revision=self._revision,
            device_map=self._device,
        )
        return model

    @override
    async def __call__(
        self,
        image_source: str | Image.Image,
        threshold: float | None = 0.3,
        tensor_format: Literal["pt"] = "pt",
    ) -> list[ImageEntity]:
        image = BaseVisionModel._get_image(image_source)
        inputs = self._processor(images=image, return_tensors=tensor_format)
        outputs = self._model(**inputs)
        target_sizes = tensor([image.size[::-1]])
        results = self._processor.post_process_object_detection(
            outputs, target_sizes=target_sizes, threshold=threshold
        )[0]

        entities = []
        for score, label, box in zip(
            results["scores"], results["labels"], results["boxes"]
        ):
            box = [round(i, 2) for i in box.tolist()]
            entities.append(
                ImageEntity(
                    label=self._model.config.id2label[label.item()],
                    score=score.item(),
                    box=box,
                )
            )
        return entities



from logging import getLogger, Logger


def _lf(items: list) -> list:
    return list(filter(None, items))


def _j(sep: str, items: list[str], *args, empty: str = "") -> str:
    real_items = _lf(items)
    return sep.join(real_items) if real_items else empty


def logger_replace(logger: Logger, logger_names: list[str]) -> None:
    assert logger and logger_names
    for logger_name in logger_names:
        updated_logger = getLogger(logger_name)
        assert updated_logger
        updated_logger.handlers = []
        for handler in logger.handlers:
            updated_logger.addHandler(handler)
        updated_logger.setLevel(logger.level)
        updated_logger.propagate = logger.propagate



from abc import abstractmethod
from ...memory import MessageMemory, MemoryStore
from ...memory.partitioner.text import TextPartition
from ...model.entities import EngineMessage, MessageRole
from ...model.nlp.sentence import SentenceTransformerModel
from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from numpy import ndarray
from typing import Literal, Optional
from ...compat import override
from uuid import UUID

Order = Literal["asc", "desc"]


class VectorFunction(StrEnum):
    # cosine distance: 1 – cosine similarity (angle between vectors). smaller
    # means  more similar
    COSINE_DISTANCE = "cosine_distance"
    # negative inner product: returns –(x·y) so that ORDER BY finds top
    # dot-product hits
    INNER_PRODUCT = "inner_product"
    # L1 (Manhattan) distance: sum of absolute differences; often more robust
    # in high-D
    L1_DISTANCE = "l1_distance"
    # L2 (Euclidean) distance: straight-line distance; smaller means more
    # similar
    L2_DISTANCE = "l2_distance"
    # Hamming distance for binary vectors
    VECTOR_DIMS = "vector_dims"
    # Jaccard distance for binary vectors
    VECTOR_NORMS = "vector_norms"


class MemoryType(StrEnum):
    CODE = "code"
    FILE = "file"
    URL = "url"
    RAW = "raw"


@dataclass(frozen=True, kw_only=True)
class Session:
    id: UUID
    agent_id: UUID
    participant_id: UUID
    messages: int
    created_at: datetime


@dataclass(frozen=True, kw_only=True)
class Memory:
    id: UUID
    model_id: str
    type: MemoryType
    participant_id: UUID
    namespace: str
    identifier: str
    data: str
    partitions: int
    symbols: dict
    created_at: datetime


@dataclass(frozen=True, kw_only=True)
class PermanentMessage:
    id: UUID
    agent_id: UUID
    model_id: str
    session_id: Optional[UUID]
    author: MessageRole
    data: str
    partitions: int
    created_at: datetime


@dataclass(frozen=True, kw_only=True)
class PermanentMessageScored(PermanentMessage):
    score: float


@dataclass(frozen=True, kw_only=True)
class PermanentMessagePartition:
    agent_id: UUID
    session_id: Optional[UUID]
    message_id: UUID
    partition: int
    data: str
    embedding: ndarray
    created_at: datetime


@dataclass(frozen=True, kw_only=True)
class PermanentMemoryPartition:
    participant_id: UUID
    memory_id: UUID
    partition: int
    data: str
    embedding: ndarray
    created_at: datetime


class PermanentMessageMemory(MessageMemory):
    _session_id: Optional[UUID] = None
    _sentence_model: SentenceTransformerModel

    def __init__(
        self,
        sentence_model: SentenceTransformerModel,
        **kwargs,
    ):
        self._sentence_model = sentence_model
        super().__init__(**kwargs)

    @property
    def has_session(self) -> bool:
        return bool(self._session_id)

    @property
    def session_id(self) -> Optional[UUID]:
        return self._session_id

    def reset(self) -> None:
        raise NotImplementedError()

    async def reset_session(self, agent_id: UUID, participant_id: UUID) -> None:
        self._session_id = await self.create_session(
            agent_id=agent_id, participant_id=participant_id
        )

    async def continue_session(
        self,
        agent_id: UUID,
        participant_id: UUID,
        session_id: UUID,
    ) -> None:
        self._session_id = await self.continue_session_and_get_id(
            agent_id=agent_id,
            participant_id=participant_id,
            session_id=session_id,
        )

    @override
    def append(self, data: EngineMessage) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def create_session(
        self, agent_id: UUID, participant_id: UUID
    ) -> UUID:
        raise NotImplementedError()

    @abstractmethod
    async def continue_session_and_get_id(
        self,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        session_id: UUID,
    ) -> UUID:
        raise NotImplementedError()

    @abstractmethod
    async def append_with_partitions(
        self,
        engine_message: EngineMessage,
        *args,
        partitions: list[TextPartition],
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def get_recent_messages(
        self,
        session_id: UUID,
        participant_id: UUID,
        *args,
        limit: Optional[int] = None,
    ) -> list[EngineMessage]:
        raise NotImplementedError()

    @abstractmethod
    async def search_messages(
        self,
        *args,
        search_partitions: list[TextPartition],
        agent_id: UUID,
        session_id: UUID,
        participant_id: UUID,
        function: VectorFunction,
        limit: Optional[int] = None,
    ) -> list[EngineMessage]:
        raise NotImplementedError()


class PermanentMemory(MemoryStore[Memory]):
    _sentence_model: SentenceTransformerModel

    def __init__(
        self, sentence_model: SentenceTransformerModel, **kwargs
    ) -> None:
        self._sentence_model = sentence_model
        super().__init__(**kwargs)

    @override
    def append(self, data: EngineMessage) -> None:
        raise NotImplementedError()

    def reset(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def append_with_partitions(
        self,
        namespace: str,
        participant_id: UUID,
        *args,
        memory_type: MemoryType,
        data: str,
        identifier: str,
        partitions: list[TextPartition],
        symbols: Optional[dict] = None,
        model_id: Optional[str] = None,
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def search_memories(
        self,
        *args,
        search_partitions: list[TextPartition],
        participant_id: UUID,
        namespace: str,
        function: VectorFunction,
        limit: Optional[int] = None,
    ) -> list[Memory]:
        raise NotImplementedError()


class RecordNotFoundException(Exception):
    def __init__(self):
        super(RecordNotFoundException, self).__init__("record_not_found")


class RecordNotSavedException(Exception):
    def __init__(self):
        super(RecordNotSavedException, self).__init__("record_not_saved")



from ast import literal_eval
from ..model.entities import ToolCall, ToolFormat
from json import JSONDecodeError, loads
from re import DOTALL, search
from typing import Any, Optional, Tuple
from xml.etree import ElementTree


class ToolCallParser:
    _eos_token: Optional[str]
    _tool_format: Optional[ToolFormat]

    def __init__(
        self,
        tool_format: Optional[ToolFormat] = None,
        eos_token: Optional[str] = None,
    ) -> None:
        self._tool_format = tool_format
        self._eos_token = eos_token

    def __call__(self, text: str) -> Optional[list[ToolCall]]:
        calls = (
            self._parse_json(text)
            if self._tool_format is ToolFormat.JSON
            else self._parse_react(text)
            if self._tool_format is ToolFormat.REACT
            else self._parse_bracket(text)
            if self._tool_format is ToolFormat.BRACKET
            else self._parse_openai_json(text)
            if self._tool_format is ToolFormat.OPENAI
            else None
        )
        if not calls:
            calls = self._parse_tag(text)
        return calls

    def set_eos_token(self, eos_token: str) -> None:
        self._eos_token = eos_token

    def _parse_json(self, text: str) -> Optional[Tuple[str, dict[str, Any]]]:
        try:
            payload = loads(text)
            return payload["tool"], payload.get("arguments", {})
        except Exception:
            return None

    def _parse_react(self, text: str) -> Optional[Tuple[str, dict[str, Any]]]:
        act = search(r"Action:\s*(\w+)", text)
        inp = search(r"Action Input:\s*({.*})", text, DOTALL)
        if act and inp:
            try:
                return act.group(1), loads(inp.group(1))
            except JSONDecodeError:
                pass
        return None

    def _parse_bracket(self, text: str) -> Optional[Tuple[str, dict[str, Any]]]:
        m = search(r"\[(\w+)\]\(([^)]+)\)", text)
        if m:
            return m.group(1), {"input": m.group(2)}
        return None

    def _parse_openai_json(
        self, text: str
    ) -> Optional[Tuple[str, dict[str, Any]]]:
        try:
            payload = loads(text)
            name = payload.get("name")
            args = payload.get("arguments", {})
            if isinstance(name, str) and isinstance(args, dict):
                return name, args
        except JSONDecodeError:
            pass
        return None

    def _parse_tag(self, text: str) -> Optional[Tuple[str, dict[str, Any]]]:
        tool_calls: list[ToolCall] = []

        if self._eos_token:
            text = text.strip().removesuffix(self._eos_token)
        try:
            root = ElementTree.fromstring(f"<root>{text}</root>")
            for element in root.findall(".//tool_call"):
                tool_call = None
                try:
                    json_text = element.text.strip()

                    try:
                        tool_call = loads(json_text)
                    except JSONDecodeError:
                        try:
                            tool_call = literal_eval(json_text)
                        except (SyntaxError, ValueError):
                            continue
                except Exception:
                    pass

                if (
                    tool_call is not None
                    and "name" in tool_call
                    and "arguments" in tool_call
                ):
                    tool_calls.append(
                        ToolCall(
                            name=tool_call["name"],
                            arguments=tool_call["arguments"],
                        )
                    )
        except ElementTree.ParseError:
            pass

        if tool_calls:
            return tool_calls

        m = search(
            r"<tool_call>\s*(\{.*?\})\s*</tool_call>", text, flags=DOTALL
        )
        if m:
            tool_call_payload = m.group(1)
            try:
                tool_call = loads(tool_call_payload)
                if (
                    tool_call
                    and "name" in tool_call
                    and "arguments" in tool_call
                ):
                    tool_calls.append(
                        ToolCall(
                            name=tool_call["name"],
                            arguments=tool_call["arguments"],
                        )
                    )
            except JSONDecodeError:
                pass

        m = search(
            r"<tool_call\s+name=\"([^\"]+)\"\s*>(\{.*?\})</tool_call>",
            text,
            DOTALL,
        )
        if m:
            try:
                tool_calls.append(
                    ToolCall(name=m.group(1), arguments=loads(m.group(2)))
                )
            except JSONDecodeError:
                pass

        m = search(
            r"<tool\s+name=\"([^\"]+)\"\s*>(\{.*?\})</tool>",
            text,
            DOTALL,
        )
        if m:
            try:
                tool_calls.append(
                    ToolCall(name=m.group(1), arguments=loads(m.group(2)))
                )
            except JSONDecodeError:
                pass

        m = search(
            r"<tool_call\s+name=\"([^\"]+)\"\s+arguments='(\{.*?\})'\s*/>",
            text,
        )
        if m:
            try:
                tool_calls.append(
                    ToolCall(name=m.group(1), arguments=loads(m.group(2)))
                )
            except JSONDecodeError:
                pass

        return tool_calls if tool_calls else None



from contextlib import nullcontext
from rich.console import Console
from rich.padding import Padding
from rich.prompt import Confirm, Prompt
from select import select
from sys import stdin
from typing import Optional


class PromptWithoutPrefix(Prompt):
    prompt_suffix = ""


class CommandAbortException(BaseException):
    pass


def confirm(console: Console, prompt: str) -> bool:
    return Confirm.ask(prompt)


def has_input(console: Console) -> bool:
    stdin_ready, __, __ = select([stdin], [], [], 0.0)
    return bool(stdin_ready)


def get_input(
    console: Console,
    prompt: Optional[str],
    *,
    echo_stdin: bool = True,
    force_prompt: bool = False,
    is_quiet: bool = False,
    prefix_line: bool = True,
    strip_prompt: bool = True,
    strip_stdin: bool = True,
    suffix_line: bool = True,
    tty_path: str = "/dev/tty",
) -> Optional[str]:
    full_prompt = f"{prompt} "
    is_input_available = has_input(console)
    if not force_prompt and is_input_available:
        input_string = stdin.read()
        if strip_stdin:
            input_string = input_string.strip()
        if prompt and not is_quiet and echo_stdin:
            console.print(
                Padding(f"{full_prompt}{input_string}", pad=(1, 0, 1, 0))
            )
    elif prompt:
        if prefix_line and not is_quiet:
            console.print("")
        try:
            with (
                open(tty_path)
                if is_input_available and force_prompt
                else nullcontext() as tty
            ):
                kwargs = {}
                if is_input_available and force_prompt:
                    kwargs["stream"] = tty
                input_string = PromptWithoutPrefix.ask(full_prompt, **kwargs)
            if strip_prompt:
                input_string = input_string.strip()
        except EOFError:
            raise CommandAbortException()
        if suffix_line and not is_quiet:
            console.print("")
    else:
        input_string = None
    return input_string



from dataclasses import dataclass
from contextlib import AsyncExitStack
from logging import Logger
from os import access, R_OK
from os.path import exists
from tomllib import load
from typing import Optional
from uuid import UUID, uuid4

from ..agent.orchestrator import Orchestrator
from ..agent.orchestrators.default import DefaultOrchestrator
from ..agent.orchestrators.json import JsonOrchestrator, Property
from ..event.manager import EventManager
from ..memory.manager import MemoryManager
from ..memory.partitioner.text import TextPartitioner
from ..model.entities import EngineUri, TransformerEngineSettings
from ..model.hubs.huggingface import HuggingfaceHub
from ..model.manager import ModelManager
from ..model.nlp.sentence import SentenceTransformerModel
from ..tool.manager import ToolManager


@dataclass(frozen=True, kw_only=True)
class OrchestratorSettings:
    agent_id: UUID
    orchestrator_type: Optional[str]
    agent_config: dict
    uri: str
    engine_config: dict
    enable_tools: Optional[object]
    call_options: Optional[dict]
    template_vars: Optional[dict]
    memory_permanent: Optional[str]
    memory_recent: bool
    sentence_model_id: str
    sentence_model_engine_config: Optional[dict]
    sentence_model_max_tokens: int
    sentence_model_overlap_size: int
    sentence_model_window_size: int
    json_config: Optional[dict]


class OrchestrationLoader:
    DEFAULT_SENTENCE_MODEL_ID = "sentence-transformers/all-MiniLM-L6-v2"

    @classmethod
    async def from_file(
        cls,
        path: str,
        *args,
        agent_id: Optional[UUID],
        hub: HuggingfaceHub,
        logger: Logger,
        participant_id: UUID,
        stack: AsyncExitStack,
        disable_memory: bool = False,
    ) -> Orchestrator:
        if not exists(path):
            raise FileNotFoundError(path)
        elif not access(path, R_OK):
            raise PermissionError(path)

        logger.debug(f"Loading agent from {path}")

        with open(path, "rb") as file:
            config = load(file)

            # Validate settings

            assert "agent" in config, "No agent section in configuration"
            assert "engine" in config, (
                "No engine section defined in configuration"
            )
            assert "uri" in config["engine"], (
                "No uri defined in engine section of configuration"
            )

            agent_config = config["agent"]
            for setting in ["role"]:
                assert setting in agent_config, (
                    f"No {setting} defined in agent section of configuration"
                )

            assert "engine" in config, (
                "No engine section defined in configuration"
            )
            assert "uri" in config["engine"], (
                "No uri defined in engine section of configuration"
            )

            uri = config["engine"]["uri"]
            engine_config = config["engine"]
            enable_tools = (
                engine_config["tools"] if "tools" in engine_config else None
            )
            engine_config.pop("uri", None)
            engine_config.pop("tools", None)
            orchestrator_type = (
                config["agent"]["type"] if "type" in config["agent"] else None
            )
            agent_id = (
                agent_id
                if agent_id
                else config["agent"]["id"]
                if "id" in config["agent"]
                else uuid4()
            )

            assert orchestrator_type is None or orchestrator_type in ["json"], (
                f"Unknown type {config['agent']['type']} in agent section "
                + "of configuration"
            )
            assert "role" in agent_config, (
                "No role defined in agent section of configuration"
            )

            call_options = config["run"] if "run" in config else None
            template_vars = config["template"] if "template" in config else None

            # Memory configuration

            memory_options = (
                config["memory"]
                if "memory" in config and not disable_memory
                else None
            )

            memory_permanent = (
                memory_options["permanent"]
                if memory_options and "permanent" in memory_options
                else None
            )
            assert not memory_permanent or isinstance(memory_permanent, str), (
                "Permanent message memory should be a string"
            )
            memory_recent = (
                memory_options["recent"]
                if memory_options and "recent" in memory_options
                else False
            )
            assert isinstance(memory_recent, bool), (
                "Recent message memory can only be set or unset"
            )

            sentence_model_id = (
                config["memory.engine"]["model_id"]
                if "memory.engine" in config
                and "model_id" in config["memory.engine"]
                else OrchestrationLoader.DEFAULT_SENTENCE_MODEL_ID
            )
            sentence_model_engine_config = (
                config["memory.engine"] if "memory.engine" in config else None
            )
            sentence_model_max_tokens = (
                config["memory.engine"]["max_tokens"]
                if sentence_model_engine_config
                and "max_tokens" in sentence_model_engine_config
                else 500
            )
            sentence_model_overlap_size = (
                config["memory.engine"]["overlap_size"]
                if sentence_model_engine_config
                and "overlap_size" in sentence_model_engine_config
                else 125
            )
            sentence_model_window_size = (
                config["memory.engine"]["window_size"]
                if sentence_model_engine_config
                and "window_size" in sentence_model_engine_config
                else 250
            )

            if sentence_model_engine_config:
                sentence_model_engine_config.pop("model_id", None)
                sentence_model_engine_config.pop("max_tokens", None)
                sentence_model_engine_config.pop("overlap_size", None)
                sentence_model_engine_config.pop("window_size", None)

            settings = OrchestratorSettings(
                agent_id=agent_id,
                orchestrator_type=orchestrator_type,
                agent_config=agent_config,
                uri=uri,
                engine_config=engine_config,
                enable_tools=enable_tools,
                call_options=call_options,
                template_vars=template_vars,
                memory_permanent=memory_permanent,
                memory_recent=memory_recent,
                sentence_model_id=sentence_model_id,
                sentence_model_engine_config=sentence_model_engine_config,
                sentence_model_max_tokens=sentence_model_max_tokens,
                sentence_model_overlap_size=sentence_model_overlap_size,
                sentence_model_window_size=sentence_model_window_size,
                json_config=config.get("json")
                if isinstance(config, dict)
                else None,
            )

            return await cls.load_from_settings(
                settings,
                hub=hub,
                logger=logger,
                participant_id=participant_id,
                stack=stack,
            )

    @classmethod
    async def load_from_settings(
        cls,
        settings: OrchestratorSettings,
        *,
        hub: HuggingfaceHub,
        logger: Logger,
        participant_id: UUID,
        stack: AsyncExitStack,
    ) -> Orchestrator:
        tool = ToolManager.create_instance(enable_tools=settings.enable_tools)

        sentence_model_engine_settings = (
            TransformerEngineSettings(**settings.sentence_model_engine_config)
            if settings.sentence_model_engine_config
            else TransformerEngineSettings()
        )

        logger.debug(
            "Loading sentence transformer model %s for agent %s",
            settings.sentence_model_id,
            settings.agent_id,
        )

        sentence_model = SentenceTransformerModel(
            model_id=settings.sentence_model_id,
            settings=sentence_model_engine_settings,
            logger=logger,
        )
        sentence_model = stack.enter_context(sentence_model)

        logger.debug(
            "Loading text partitioner for model %s for agent %s with settings (%s, %s, %s)",
            settings.sentence_model_id,
            settings.agent_id,
            settings.sentence_model_max_tokens,
            settings.sentence_model_overlap_size,
            settings.sentence_model_window_size,
        )

        text_partitioner = TextPartitioner(
            model=sentence_model,
            logger=logger,
            max_tokens=settings.sentence_model_max_tokens,
            overlap_size=settings.sentence_model_overlap_size,
            window_size=settings.sentence_model_window_size,
        )

        logger.debug("Loading memory manager for agent %s", settings.agent_id)

        memory = await MemoryManager.create_instance(
            agent_id=settings.agent_id,
            participant_id=participant_id,
            text_partitioner=text_partitioner,
            with_permanent_message_memory=settings.memory_permanent,
            with_recent_message_memory=settings.memory_recent,
        )

        event_manager = EventManager()
        event_manager.add_listener(
            lambda e: logger.debug(f"Event {e.type}: {e.payload}")
        )

        logger.debug(
            "Creating orchestrator %s #%s",
            settings.orchestrator_type,
            settings.agent_id,
        )

        model_manager = ModelManager(hub, logger)
        model_manager = stack.enter_context(model_manager)

        engine_uri = model_manager.parse_uri(settings.uri)
        engine_settings = model_manager.get_engine_settings(
            engine_uri,
            settings=settings.engine_config,
        )

        if settings.orchestrator_type == "json":
            assert settings.json_config is not None
            agent = cls._load_json_orchestrator(
                engine_uri=engine_uri,
                engine_settings=engine_settings,
                logger=logger,
                model_manager=model_manager,
                memory=memory,
                tool=tool,
                event_manager=event_manager,
                config={"json": settings.json_config},
                agent_config=settings.agent_config,
                call_options=settings.call_options,
                template_vars=settings.template_vars,
            )
        else:
            agent = DefaultOrchestrator(
                engine_uri,
                logger,
                model_manager,
                memory,
                tool,
                event_manager,
                name=settings.agent_config.get("name"),
                role=settings.agent_config["role"],
                task=settings.agent_config.get("task"),
                instructions=settings.agent_config.get("instructions"),
                rules=settings.agent_config.get("rules"),
                settings=engine_settings,
                call_options=settings.call_options,
                template_vars=settings.template_vars,
            )

        return agent

    @staticmethod
    def _load_json_orchestrator(
        engine_uri: EngineUri,
        engine_settings: TransformerEngineSettings,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        config: dict,
        agent_config: dict,
        call_options: Optional[dict],
        template_vars: Optional[dict],
    ) -> JsonOrchestrator:
        assert "json" in config, "No json section in configuration"
        assert "instructions" in agent_config, (
            "No instructions defined in agent section of configuration"
        )
        assert "task" in agent_config, (
            "No task defined in agent section of configuration"
        )
        assert "role" in agent_config, (
            "No role defined in agent section of configuration"
        )

        properties: list[Property] = []
        for property_name in config.get("json", []):
            output_property = config["json"][property_name]
            properties.append(
                Property(
                    name=property_name,
                    data_type=output_property["type"],
                    description=output_property["description"],
                )
            )

        assert properties, "No properties defined in configuration"

        agent = JsonOrchestrator(
            engine_uri,
            logger,
            model_manager,
            memory,
            tool,
            event_manager,
            properties,
            name=agent_config["name"] if "name" in agent_config else None,
            role=agent_config["role"],
            task=agent_config["task"],
            instructions=agent_config["instructions"],
            rules=agent_config["rules"] if "rules" in agent_config else None,
            settings=engine_settings,
            call_options=call_options,
            template_vars=template_vars,
        )
        return agent
